#
# PySNMP MIB module BFD-STD-MIB (http://snmplabs.com/pysmi)
# ASN.1 source file://./BFD-STD-MIB.mib
# Produced by pysmi-1.1.10 at Tue Oct 10 14:11:21 2023
# On host simtve-ThinkPad-X1-Carbon-Gen-9 platform Linux version 5.15.0-84-generic by user simon
# Using Python version 3.9.18 (main, Aug 25 2023, 13:20:04) 
#
Integer, ObjectIdentifier, OctetString = mibBuilder.importSymbols("ASN1", "Integer", "ObjectIdentifier", "OctetString")
NamedValues, = mibBuilder.importSymbols("ASN1-ENUMERATION", "NamedValues")
ConstraintsUnion, ValueSizeConstraint, SingleValueConstraint, ConstraintsIntersection, ValueRangeConstraint = mibBuilder.importSymbols("ASN1-REFINEMENT", "ConstraintsUnion", "ValueSizeConstraint", "SingleValueConstraint", "ConstraintsIntersection", "ValueRangeConstraint")
InetAddressType, InetPortNumber, InetAddress = mibBuilder.importSymbols("INET-ADDRESS-MIB", "InetAddressType", "InetPortNumber", "InetAddress")
jnxBfdExperiment, = mibBuilder.importSymbols("JUNIPER-EXPERIMENT-MIB", "jnxBfdExperiment")
ObjectGroup, ModuleCompliance, NotificationGroup = mibBuilder.importSymbols("SNMPv2-CONF", "ObjectGroup", "ModuleCompliance", "NotificationGroup")
Counter64, MibIdentifier, Counter32, Gauge32, IpAddress, Bits, ObjectIdentity, iso, NotificationType, TimeTicks, Unsigned32, MibScalar, MibTable, MibTableRow, MibTableColumn, Integer32, ModuleIdentity = mibBuilder.importSymbols("SNMPv2-SMI", "Counter64", "MibIdentifier", "Counter32", "Gauge32", "IpAddress", "Bits", "ObjectIdentity", "iso", "NotificationType", "TimeTicks", "Unsigned32", "MibScalar", "MibTable", "MibTableRow", "MibTableColumn", "Integer32", "ModuleIdentity")
DisplayString, StorageType, RowStatus, TruthValue, TimeStamp, TextualConvention = mibBuilder.importSymbols("SNMPv2-TC", "DisplayString", "StorageType", "RowStatus", "TruthValue", "TimeStamp", "TextualConvention")
bfdMIB = ModuleIdentity((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1))
bfdMIB.setRevisions(('2005-08-22 12:00',))

if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    if mibBuilder.loadTexts: bfdMIB.setRevisionsDescriptions(('Initial version. Published as RFC xxxx.',))
if mibBuilder.loadTexts: bfdMIB.setLastUpdated('200507221200Z')
if mibBuilder.loadTexts: bfdMIB.setOrganization('IETF')
if mibBuilder.loadTexts: bfdMIB.setContactInfo(' Thomas D. Nadeau Cisco Systems, Inc. Email: tnadeau@cisco.com Zafar Ali Cisco Systems, Inc. Email: zali@cisco.com ')
if mibBuilder.loadTexts: bfdMIB.setDescription('Bidirectional Forwarding Management Information Base.')
bfdNotifications = MibIdentifier((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 0))
bfdObjects = MibIdentifier((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1))
bfdConformance = MibIdentifier((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 3))
bfdScalarObjects = MibIdentifier((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 1))
class BfdSessIndexTC(TextualConvention, Unsigned32):
    description = 'An index used to uniquely identify BFD sessions.'
    status = 'current'
    displayHint = 'd'
    subtypeSpec = Unsigned32.subtypeSpec + ValueRangeConstraint(1, 4294967295)

class BfdInterval(TextualConvention, Unsigned32):
    description = 'The BFD interval delay in microseconds.'
    status = 'current'
    subtypeSpec = Unsigned32.subtypeSpec + ValueRangeConstraint(1, 4294967295)

class BfdDiag(TextualConvention, Integer32):
    description = 'A common BFD diagnostic code.'
    status = 'current'
    subtypeSpec = Integer32.subtypeSpec + ConstraintsUnion(SingleValueConstraint(1, 2, 3, 4, 5, 6, 7, 8, 9))
    namedValues = NamedValues(("noDiagnostic", 1), ("controlDetectionTimeExpired", 2), ("echoFunctionFailed", 3), ("neighborSignaledSessionDown", 4), ("forwardingPlaneReset", 5), ("pathDown", 6), ("concatenatedPathDown", 7), ("administrativelyDown", 8), ("reverseConcatenatedPathDown", 9))

bfdAdminStatus = MibScalar((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 1, 1), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(1, 2))).clone(namedValues=NamedValues(("enabled", 1), ("disabled", 2))).clone('enabled')).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdAdminStatus.setStatus('current')
if mibBuilder.loadTexts: bfdAdminStatus.setDescription("The global administrative status of BFD in this router. The value 'enabled' denotes that the BFD Process is active on at least one interface; 'disabled' disables it on all interfaces.")
bfdVersionNumber = MibScalar((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 1, 3), Unsigned32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdVersionNumber.setReference(' BFD Version 0 (draft-katz-ward-bfd-02.txt)')
if mibBuilder.loadTexts: bfdVersionNumber.setStatus('current')
if mibBuilder.loadTexts: bfdVersionNumber.setDescription('The current version number of the BFD protocol.')
bfdSessTable = MibTable((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 2), )
if mibBuilder.loadTexts: bfdSessTable.setReference('BFD Version 0 (draft-katz-ward-bfd-02.txt)')
if mibBuilder.loadTexts: bfdSessTable.setStatus('current')
if mibBuilder.loadTexts: bfdSessTable.setDescription('The BFD Session Table describes the BFD sessions.')
bfdSessEntry = MibTableRow((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 2, 1), ).setIndexNames((0, "BFD-STD-MIB", "bfdSessIndex"))
if mibBuilder.loadTexts: bfdSessEntry.setStatus('current')
if mibBuilder.loadTexts: bfdSessEntry.setDescription('The BFD Session Entry describes BFD session.')
bfdSessIndex = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 2, 1, 1), BfdSessIndexTC())
if mibBuilder.loadTexts: bfdSessIndex.setStatus('current')
if mibBuilder.loadTexts: bfdSessIndex.setDescription('This object contains an index used to represent a unique BFD session on this device.')
bfdSessApplicationId = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 2, 1, 2), Unsigned32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessApplicationId.setStatus('current')
if mibBuilder.loadTexts: bfdSessApplicationId.setDescription('This object contains an index used to indicate a local application which owns or maintains this BFD session. For instance, the MPLS VPN process may maintain a subset of the total number of BFD sessions. This application ID provides a convenient way to segregate sessions by the applications which maintain them.')
bfdSessDiscriminator = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 2, 1, 3), Unsigned32().subtype(subtypeSpec=ValueRangeConstraint(1, 4294967295))).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessDiscriminator.setStatus('current')
if mibBuilder.loadTexts: bfdSessDiscriminator.setDescription('This object specifies the local discriminator for this BFD session, used to uniquely identify it.')
bfdSessRemoteDiscr = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 2, 1, 4), Unsigned32().subtype(subtypeSpec=ValueRangeConstraint(1, 4294967295))).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessRemoteDiscr.setStatus('current')
if mibBuilder.loadTexts: bfdSessRemoteDiscr.setDescription('This object specifies the session discriminator chosen by the remote system for this BFD session.')
bfdSessUdpPort = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 2, 1, 5), InetPortNumber()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessUdpPort.setReference('draft-katz-ward-bfd-02.txt and draft-raggarwa-mpls-bfd-00.txt')
if mibBuilder.loadTexts: bfdSessUdpPort.setStatus('current')
if mibBuilder.loadTexts: bfdSessUdpPort.setDescription('The UDP Port for BFD. The default value is the well-known value for this port.')
bfdSessState = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 2, 1, 6), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(1, 2, 3, 4))).clone(namedValues=NamedValues(("adminDown", 1), ("down", 2), ("init", 3), ("up", 4)))).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessState.setStatus('current')
if mibBuilder.loadTexts: bfdSessState.setDescription('The perceived state of the BFD session.')
bfdSessRemoteHeardFlag = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 2, 1, 7), TruthValue()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessRemoteHeardFlag.setStatus('current')
if mibBuilder.loadTexts: bfdSessRemoteHeardFlag.setDescription('This object specifies status of BFD packet reception from the remote system. Specifically, it is set to true(1) if the local system is actively receiving BFD packets from the remote system, and is set to false(0) if the local system has not received BFD packets recently (within the detection time) or if the local system is attempting to tear down the BFD session.')
bfdSessDiag = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 2, 1, 8), BfdDiag()).setMaxAccess("accessiblefornotify")
if mibBuilder.loadTexts: bfdSessDiag.setStatus('current')
if mibBuilder.loadTexts: bfdSessDiag.setDescription("A diagnostic code specifying the local system's reason for the last transition of the session from up(1) to some other state.")
bfdSessOperMode = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 2, 1, 9), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(1, 2, 3, 4))).clone(namedValues=NamedValues(("asyncModeWEchoFun", 1), ("asynchModeWOEchoFun", 2), ("demandModeWEchoFunction", 3), ("demandModeWOEchoFunction", 4)))).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessOperMode.setStatus('current')
if mibBuilder.loadTexts: bfdSessOperMode.setDescription('This object specifies current operating mode that BFD session is operating in. A value of AsyncModeWEchoFun(1) ... A value of AsynchModeWOEchoFun(2) ... A value of DemandModeWEchoFunction(3) ... A value of DemandModeWOEchoFunction(4) ... ')
bfdSessDemandModeDesiredFlag = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 2, 1, 10), TruthValue().clone('false')).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessDemandModeDesiredFlag.setStatus('current')
if mibBuilder.loadTexts: bfdSessDemandModeDesiredFlag.setDescription("This object indicates that the local system's desire to use Demand mode. Specifically, it is set to true(1) if the local system wishes to use Demand mode or false(0) if not")
bfdSessEchoFuncModeDesiredFlag = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 2, 1, 11), TruthValue().clone('false')).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessEchoFuncModeDesiredFlag.setStatus('current')
if mibBuilder.loadTexts: bfdSessEchoFuncModeDesiredFlag.setDescription("This object indicates that the local system's desire to use Echo mode. Specifically, it is set to true(1) if the local system wishes to use Echo mode or false(0) if not")
bfdSessControlPlanIndepFlag = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 2, 1, 12), TruthValue().clone('false')).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessControlPlanIndepFlag.setStatus('current')
if mibBuilder.loadTexts: bfdSessControlPlanIndepFlag.setDescription("This object indicates that the local system's ability to continue to function through a disruption of the control plane. Specifically, it is set to true(1) if the local system BFD implementation is independent of the control plane. Otherwise, the value is set to false(0)")
bfdSessAddrType = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 2, 1, 13), InetAddressType()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessAddrType.setStatus('current')
if mibBuilder.loadTexts: bfdSessAddrType.setDescription('This object specifies IP address of the interface associated with this BFD session. Only values unknown(0), ipv4(1) or ipv6(2) have to be supported. A value of unknown(0) is allowed only when the outgoing interface is of type point-to-point, or when the BFD session is not associated with a specific interface. If any other unsupported values are attempted in a set operation, the agent MUST return an inconsistentValue error. ')
bfdSessAddr = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 2, 1, 14), InetAddress()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessAddr.setStatus('current')
if mibBuilder.loadTexts: bfdSessAddr.setDescription('This object specifies IP address of the interface associated with this BFD session. It can also be used to enabled BFD on a specific interface. The value is set to zero when BFD session is not associated with a specific interface. ')
bfdSessDesiredMinTxInterval = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 2, 1, 15), BfdInterval()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessDesiredMinTxInterval.setStatus('current')
if mibBuilder.loadTexts: bfdSessDesiredMinTxInterval.setDescription('This object specifies the minimum interval, in microseconds, that the local system would like to use when transmitting BFD Control packets.')
bfdSessDesiredMinRxInterval = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 2, 1, 16), BfdInterval()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessDesiredMinRxInterval.setStatus('current')
if mibBuilder.loadTexts: bfdSessDesiredMinRxInterval.setDescription('This object specifies the minimum interval, in microseconds, between received BFD Control packets the local system is capable of supporting.')
bfdSessDesiredMinEchoRxInterval = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 2, 1, 17), BfdInterval()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessDesiredMinEchoRxInterval.setStatus('current')
if mibBuilder.loadTexts: bfdSessDesiredMinEchoRxInterval.setDescription('This object specifies the minimum interval, in microseconds, between received BFD Echo packets that this system is capable of supporting.')
bfdSessDetectMult = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 2, 1, 18), Unsigned32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessDetectMult.setStatus('current')
if mibBuilder.loadTexts: bfdSessDetectMult.setDescription('This object specifies the Detect time multiplier.')
bfdSessStorType = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 2, 1, 19), StorageType()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessStorType.setStatus('current')
if mibBuilder.loadTexts: bfdSessStorType.setDescription("This variable indicates the storage type for this object. Conceptual rows having the value 'permanent' need not allow write-access to any columnar objects in the row.")
bfdSessRowStatus = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 2, 1, 20), RowStatus()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessRowStatus.setStatus('current')
if mibBuilder.loadTexts: bfdSessRowStatus.setDescription('This variable is used to create, modify, and/or delete a row in this table. When a row in this table has a row in the active(1) state, no objects in this row can be modified except the bfdSessRowStatus and bfdSessStorageType.')
bfdSessAuthPresFlag = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 2, 1, 21), TruthValue().clone('false')).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessAuthPresFlag.setStatus('current')
if mibBuilder.loadTexts: bfdSessAuthPresFlag.setDescription("This object indicates that the local system's desire to use Authentication. Specifically, it is set to true(1) if the local system wishes the session to be authenticated or false(0) if not")
bfdSessAuthenticationType = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 2, 1, 22), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(1, 2, 3, 4, 5))).clone(namedValues=NamedValues(("simplePassword", 1), ("keyedMD5", 2), ("meticulousKeyedMD5", 3), ("keyedSHA1", 4), ("meticulousKeyedSHA1", 5)))).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessAuthenticationType.setStatus('current')
if mibBuilder.loadTexts: bfdSessAuthenticationType.setDescription('The Authentication Type used for this BFD session. This field is valid only when the Authentication Present bit is set')
bfdSessPerfTable = MibTable((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 3), )
if mibBuilder.loadTexts: bfdSessPerfTable.setStatus('current')
if mibBuilder.loadTexts: bfdSessPerfTable.setDescription('This table specifies BFD Session performance counters.')
bfdSessPerfEntry = MibTableRow((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 3, 1), )
bfdSessEntry.registerAugmentions(("BFD-STD-MIB", "bfdSessPerfEntry"))
bfdSessPerfEntry.setIndexNames(*bfdSessEntry.getIndexNames())
if mibBuilder.loadTexts: bfdSessPerfEntry.setStatus('current')
if mibBuilder.loadTexts: bfdSessPerfEntry.setDescription('An entry in this table is created by a BFD-enabled node for every BFD Session. bfdCounterDiscontinuityTime is used to indicate potential discontinuity for all counter objects in this table.')
bfdSessPerfPktIn = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 3, 1, 1), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessPerfPktIn.setStatus('current')
if mibBuilder.loadTexts: bfdSessPerfPktIn.setDescription('The total number of BFD messages received for this BFD session.')
bfdSessPerfPktOut = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 3, 1, 2), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessPerfPktOut.setStatus('current')
if mibBuilder.loadTexts: bfdSessPerfPktOut.setDescription('The total number of BFD messages sent for this BFD session.')
bfdSessUpTime = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 3, 1, 3), TimeStamp()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessUpTime.setStatus('current')
if mibBuilder.loadTexts: bfdSessUpTime.setDescription('The value of sysUpTime on the most recent occasion at which the session came up. If no such up event exists this object contains a zero value.')
bfdSessPerfLastSessDownTime = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 3, 1, 4), TimeStamp()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessPerfLastSessDownTime.setStatus('current')
if mibBuilder.loadTexts: bfdSessPerfLastSessDownTime.setDescription('The value of sysUpTime on the most recent occasion at which the last time communication was lost with the neighbor. If no such down event exist this object contains a zero value.')
bfdSessPerfLastCommLostDiag = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 3, 1, 5), BfdDiag()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessPerfLastCommLostDiag.setStatus('current')
if mibBuilder.loadTexts: bfdSessPerfLastCommLostDiag.setDescription('The BFD diag code for the last time communication was lost with the neighbor. If no such down event exists this object contains a zero value.')
bfdSessPerfSessUpCount = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 3, 1, 6), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessPerfSessUpCount.setStatus('current')
if mibBuilder.loadTexts: bfdSessPerfSessUpCount.setDescription('The number of times this session has gone into the Up state since the router last rebooted.')
bfdSessPerfDiscTime = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 3, 1, 7), TimeStamp()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessPerfDiscTime.setStatus('current')
if mibBuilder.loadTexts: bfdSessPerfDiscTime.setDescription('The value of sysUpTime on the most recent occasion at which any one or more of the session counters suffered a discontinuity. The relevant counters are the specific instances associated with this BFD session of any Counter32 object contained in the BfdSessPerfTable. If no such discontinuities have occurred since the last re-initialization of the local management subsystem, then this object contains a zero value.')
bfdSessPerfPktInHC = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 3, 1, 8), Counter64()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessPerfPktInHC.setStatus('current')
if mibBuilder.loadTexts: bfdSessPerfPktInHC.setDescription('This value represents the total number of BFD messages received for this BFD session. It MUST be equal to the least significant 32 bits of bfdSessPerfPktIn if bfdSessPerfPktInHC is supported according to the rules spelled out in RFC2863.')
bfdSessPerfPktOutHC = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 3, 1, 9), Counter64()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessPerfPktOutHC.setStatus('current')
if mibBuilder.loadTexts: bfdSessPerfPktOutHC.setDescription('This value represents the total number of total number of BFD messages transmitted for this BFD session. It MUST be equal to the least significant 32 bits of bfdSessPerfPktIn if bfdSessPerfPktOutHC is supported according to the rules spelled out in RFC2863.')
bfdSessMapTable = MibTable((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 4), )
if mibBuilder.loadTexts: bfdSessMapTable.setReference('BFD Version 0 (draft-katz-ward-bfd-02.txt)')
if mibBuilder.loadTexts: bfdSessMapTable.setStatus('current')
if mibBuilder.loadTexts: bfdSessMapTable.setDescription('The BFD Session Mapping Table maps the complex indexing of the BFD sessions to the flat BFDIndex used in the BfdSessionTable. Implementors need to be aware that if the value of the bfdSessAddr (an OID) has more that 111 sub-identifiers, then OIDs of column instances in this table will have more than 128 sub-identifiers and cannot be accessed using SNMPv1, SNMPv2c, or SNMPv3. ')
bfdSessMapEntry = MibTableRow((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 4, 1), ).setIndexNames((0, "BFD-STD-MIB", "bfdSessApplicationId"), (0, "BFD-STD-MIB", "bfdSessDiscriminator"), (0, "BFD-STD-MIB", "bfdSessAddrType"), (0, "BFD-STD-MIB", "bfdSessAddr"))
if mibBuilder.loadTexts: bfdSessMapEntry.setStatus('current')
if mibBuilder.loadTexts: bfdSessMapEntry.setDescription('The BFD Session Entry describes BFD session that is mapped to this index. Implementors need to be aware that if the value of the mplsInSegmentMapLabelPtrIndex (an OID) has more that 111 sub-identifiers, then OIDs of column instances in this table will have more than 128 sub-identifiers and cannot be accessed using SNMPv1, SNMPv2c, or SNMPv3.')
bfdSessMapBfdIndex = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 4, 1, 1), BfdSessIndexTC()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessMapBfdIndex.setStatus('current')
if mibBuilder.loadTexts: bfdSessMapBfdIndex.setDescription('This object specifies the BfdIndex referred to by the indexes of this row. In essence, a mapping is provided between these indexes and the BfdSessTable.')
bfdSessNotificationsEnable = MibScalar((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 1, 1, 4), TruthValue().clone('false')).setMaxAccess("readonly")
if mibBuilder.loadTexts: bfdSessNotificationsEnable.setReference('See also RFC3413 for explanation that notifications are under the ultimate control of the MIB modules in this document.')
if mibBuilder.loadTexts: bfdSessNotificationsEnable.setStatus('current')
if mibBuilder.loadTexts: bfdSessNotificationsEnable.setDescription('If this object is set to true(1), then it enables the emission of bfdSessUp and bfdSessDown notifications; otherwise these notifications are not emitted.')
bfdSessUp = NotificationType((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 0, 1)).setObjects(("BFD-STD-MIB", "bfdSessDiag"), ("BFD-STD-MIB", "bfdSessDiag"))
if mibBuilder.loadTexts: bfdSessUp.setStatus('current')
if mibBuilder.loadTexts: bfdSessUp.setDescription('This notification is generated when the bfdSessState object for one or more contiguous entries in bfdSessTable are about to enter the up(2) state from some other state. The included values of bfdSessDiag MUST both be set equal to this new state (i.e: up(1)). The two instances of bfdSessDiag in this notification indicate the range of indexes that are affected. Note that all the indexes of the two ends of the range can be derived from the instance identifiers of these two objects. For the cases where a contiguous range of sessions have transitioned into the up(1) state at roughly the same time, the device SHOULD issue a single notification for each range of contiguous indexes in an effort to minimize the emission of a large number of notifications. If a notification has to be issued for just a single bfdSessEntry, then the instance identifier (and values) of the two bfdSessDiag objects MUST be the identical.')
bfdSessDown = NotificationType((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 0, 2)).setObjects(("BFD-STD-MIB", "bfdSessDiag"), ("BFD-STD-MIB", "bfdSessDiag"))
if mibBuilder.loadTexts: bfdSessDown.setStatus('current')
if mibBuilder.loadTexts: bfdSessDown.setDescription('This notification is generated when the bfdSessState object for one or more contiguous entries in bfdSessTable are about to enter the down(4) or adminDown(5) states from some other state. The included values of bfdSessDiag MUST both be set equal to this new state (i.e: down(4) or adminDown(5)). The two instances of bfdSessDiag in this notification indicate the range of indexes that are affected. Note that all the indexes of the two ends of the range can be derived from the instance identifiers of these two objects. For cases where a contiguous range of sessions have transitioned into the down(4) or adminDown(5) states at roughly the same time, the device SHOULD issue a single notification for each range of contiguous indexes in an effort to minimize the emission of a large number of notifications. If a notification has to be issued for just a single bfdSessEntry, then the instance identifier (and values) of the two bfdSessDiag objects MUST be the identical.')
bfdGroups = MibIdentifier((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 3, 1))
bfdCompliances = MibIdentifier((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 3, 2))
bfdModuleFullCompliance = ModuleCompliance((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 3, 2, 1)).setObjects(("BFD-STD-MIB", "bfdSessionGroup"), ("BFD-STD-MIB", "bfdSessionPerfGroup"), ("BFD-STD-MIB", "bfdSessionPerfHCGroup"), ("BFD-STD-MIB", "bfdNotificationGroup"))

if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    bfdModuleFullCompliance = bfdModuleFullCompliance.setStatus('current')
if mibBuilder.loadTexts: bfdModuleFullCompliance.setDescription('Compliance statement for agents that provide full support for BFD-MIB. Such devices can then be monitored and also be configured using this MIB module.')
bfdSessionGroup = ObjectGroup((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 3, 1, 1)).setObjects(("BFD-STD-MIB", "bfdSessNotificationsEnable"), ("BFD-STD-MIB", "bfdAdminStatus"), ("BFD-STD-MIB", "bfdVersionNumber"), ("BFD-STD-MIB", "bfdSessApplicationId"), ("BFD-STD-MIB", "bfdSessDiscriminator"), ("BFD-STD-MIB", "bfdSessAddrType"), ("BFD-STD-MIB", "bfdSessAddr"), ("BFD-STD-MIB", "bfdSessRemoteDiscr"), ("BFD-STD-MIB", "bfdSessUdpPort"), ("BFD-STD-MIB", "bfdSessState"), ("BFD-STD-MIB", "bfdSessRemoteHeardFlag"), ("BFD-STD-MIB", "bfdSessDiag"), ("BFD-STD-MIB", "bfdSessOperMode"), ("BFD-STD-MIB", "bfdSessDemandModeDesiredFlag"), ("BFD-STD-MIB", "bfdSessEchoFuncModeDesiredFlag"), ("BFD-STD-MIB", "bfdSessControlPlanIndepFlag"), ("BFD-STD-MIB", "bfdSessDesiredMinTxInterval"), ("BFD-STD-MIB", "bfdSessDesiredMinRxInterval"), ("BFD-STD-MIB", "bfdSessDesiredMinEchoRxInterval"), ("BFD-STD-MIB", "bfdSessDetectMult"), ("BFD-STD-MIB", "bfdSessStorType"), ("BFD-STD-MIB", "bfdSessRowStatus"), ("BFD-STD-MIB", "bfdSessMapBfdIndex"), ("BFD-STD-MIB", "bfdSessAuthPresFlag"), ("BFD-STD-MIB", "bfdSessAuthenticationType"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    bfdSessionGroup = bfdSessionGroup.setStatus('current')
if mibBuilder.loadTexts: bfdSessionGroup.setDescription('Collection of objects needed for BFD sessions.')
bfdSessionPerfGroup = ObjectGroup((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 3, 1, 2)).setObjects(("BFD-STD-MIB", "bfdSessPerfPktIn"), ("BFD-STD-MIB", "bfdSessPerfPktOut"), ("BFD-STD-MIB", "bfdSessUpTime"), ("BFD-STD-MIB", "bfdSessPerfLastSessDownTime"), ("BFD-STD-MIB", "bfdSessPerfLastCommLostDiag"), ("BFD-STD-MIB", "bfdSessPerfSessUpCount"), ("BFD-STD-MIB", "bfdSessPerfDiscTime"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    bfdSessionPerfGroup = bfdSessionPerfGroup.setStatus('current')
if mibBuilder.loadTexts: bfdSessionPerfGroup.setDescription('Collection of objects needed to monitor the performance of BFD sessions.')
bfdSessionPerfHCGroup = ObjectGroup((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 3, 1, 3)).setObjects(("BFD-STD-MIB", "bfdSessPerfPktInHC"), ("BFD-STD-MIB", "bfdSessPerfPktOutHC"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    bfdSessionPerfHCGroup = bfdSessionPerfHCGroup.setStatus('current')
if mibBuilder.loadTexts: bfdSessionPerfHCGroup.setDescription('Collection of objects needed to monitor the performance of BFD sessions for which the values of bfdSessPerfPktIn, bfdSessPerfPktOut wrap around too quickly.')
bfdNotificationGroup = NotificationGroup((1, 3, 6, 1, 4, 1, 2636, 5, 3, 1, 3, 1, 4)).setObjects(("BFD-STD-MIB", "bfdSessUp"), ("BFD-STD-MIB", "bfdSessDown"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    bfdNotificationGroup = bfdNotificationGroup.setStatus('current')
if mibBuilder.loadTexts: bfdNotificationGroup.setDescription('Set of notifications implemented in this module.')
mibBuilder.exportSymbols("BFD-STD-MIB", bfdSessDown=bfdSessDown, bfdSessRemoteHeardFlag=bfdSessRemoteHeardFlag, bfdSessPerfPktOut=bfdSessPerfPktOut, BfdSessIndexTC=BfdSessIndexTC, bfdSessPerfEntry=bfdSessPerfEntry, bfdGroups=bfdGroups, bfdVersionNumber=bfdVersionNumber, bfdScalarObjects=bfdScalarObjects, bfdSessRemoteDiscr=bfdSessRemoteDiscr, bfdSessPerfLastCommLostDiag=bfdSessPerfLastCommLostDiag, bfdNotifications=bfdNotifications, bfdSessDetectMult=bfdSessDetectMult, bfdConformance=bfdConformance, bfdSessPerfPktIn=bfdSessPerfPktIn, bfdSessPerfDiscTime=bfdSessPerfDiscTime, bfdModuleFullCompliance=bfdModuleFullCompliance, bfdSessPerfPktOutHC=bfdSessPerfPktOutHC, bfdSessMapBfdIndex=bfdSessMapBfdIndex, bfdSessDiag=bfdSessDiag, bfdSessNotificationsEnable=bfdSessNotificationsEnable, bfdSessUp=bfdSessUp, bfdSessAuthenticationType=bfdSessAuthenticationType, bfdSessTable=bfdSessTable, bfdSessMapEntry=bfdSessMapEntry, bfdAdminStatus=bfdAdminStatus, bfdSessIndex=bfdSessIndex, bfdSessDesiredMinTxInterval=bfdSessDesiredMinTxInterval, bfdSessAddrType=bfdSessAddrType, bfdSessState=bfdSessState, bfdSessionGroup=bfdSessionGroup, bfdSessStorType=bfdSessStorType, bfdSessDiscriminator=bfdSessDiscriminator, bfdSessOperMode=bfdSessOperMode, bfdNotificationGroup=bfdNotificationGroup, bfdSessAuthPresFlag=bfdSessAuthPresFlag, BfdInterval=BfdInterval, bfdSessPerfLastSessDownTime=bfdSessPerfLastSessDownTime, bfdSessDesiredMinEchoRxInterval=bfdSessDesiredMinEchoRxInterval, bfdCompliances=bfdCompliances, bfdSessionPerfGroup=bfdSessionPerfGroup, bfdSessDemandModeDesiredFlag=bfdSessDemandModeDesiredFlag, bfdSessUdpPort=bfdSessUdpPort, bfdObjects=bfdObjects, bfdSessPerfTable=bfdSessPerfTable, bfdSessApplicationId=bfdSessApplicationId, bfdSessionPerfHCGroup=bfdSessionPerfHCGroup, bfdSessEchoFuncModeDesiredFlag=bfdSessEchoFuncModeDesiredFlag, bfdMIB=bfdMIB, BfdDiag=BfdDiag, bfdSessEntry=bfdSessEntry, bfdSessDesiredMinRxInterval=bfdSessDesiredMinRxInterval, bfdSessPerfSessUpCount=bfdSessPerfSessUpCount, bfdSessAddr=bfdSessAddr, bfdSessRowStatus=bfdSessRowStatus, bfdSessPerfPktInHC=bfdSessPerfPktInHC, bfdSessMapTable=bfdSessMapTable, PYSNMP_MODULE_ID=bfdMIB, bfdSessControlPlanIndepFlag=bfdSessControlPlanIndepFlag, bfdSessUpTime=bfdSessUpTime)



from datetime import timedelta

import pytest

from zino.statemodels import ReachabilityEvent, ReachabilityState
from zino.tasks.errors import DeviceUnreachableError
from zino.time import now


class TestReachableTask:
    async def test_given_event_creation_config_of_false_run_should_not_create_event_if_device_is_reachable(
        self, reachable_task
    ):
        task = reachable_task
        task.config.event.make_events_for_new_devices = False
        assert (await task.run()) is None
        event = task.state.events.get(task.device.name, None, ReachabilityEvent)
        assert not event

    async def test_given_event_creation_config_of_true_run_should_create_event_if_device_is_new_and_reachable(
        self, reachable_task
    ):
        task = reachable_task
        task.config.event.make_events_for_new_devices = True
        assert (await task.run()) is None
        event = task.state.events.get(task.device.name, None, ReachabilityEvent)
        assert event
        assert f"New device: {reachable_task.device.name} reachable" in [log.message for log in event.log]

    async def test_run_should_create_event_if_device_is_known_and_unreachable(self, unreachable_task):
        task = unreachable_task
        task.state.devices[task.device.name].reachable_in_last_run = True
        with pytest.raises(DeviceUnreachableError):
            await task.run()
        event = task.state.events.get(task.device.name, None, ReachabilityEvent)
        assert event
        assert f"{unreachable_task.device.name} no-response" in [log.message for log in event.log]

    async def test_run_should_create_event_if_device_is_new_and_unreachable(self, unreachable_task):
        task = unreachable_task
        task.config.event.make_events_for_new_devices = True
        with pytest.raises(DeviceUnreachableError):
            await task.run()
        event = task.state.events.get(task.device.name, None, ReachabilityEvent)
        assert event
        assert f"New device: {unreachable_task.device.name} no-response" in [log.message for log in event.log]

    async def test_run_should_update_device_status_to_reachable_in_last_run_false_when_device_is_unreachable(
        self, unreachable_task
    ):
        task = unreachable_task
        with pytest.raises(DeviceUnreachableError):
            await task.run()
        assert task.state.devices[task.device.name].reachable_in_last_run is False

    async def test_run_should_start_extra_job_if_device_is_unreachable(self, unreachable_task):
        with pytest.raises(DeviceUnreachableError):
            await unreachable_task.run()
        assert unreachable_task._extra_job_is_running()

    async def test_run_should_not_start_extra_job_if_device_is_reachable(self, reachable_task):
        task = reachable_task
        assert (await task.run()) is None
        assert not task._extra_job_is_running()

    async def test_run_should_update_event_to_reachable_when_device_is_reachable(self, reachable_task):
        task = reachable_task
        event = task.state.events.create_event(task.device.name, None, ReachabilityEvent)
        event.reachability = ReachabilityState.NORESPONSE
        event.lasttrans = now() - timedelta(minutes=5)
        task.state.events.commit(event)

        assert (await task.run()) is None
        updated_event = task.state.events[event.id]
        assert updated_event.reachability == ReachabilityState.REACHABLE

    async def test_run_should_update_device_status_to_reachable_in_last_run_true_when_device_is_reachable(
        self, reachable_task
    ):
        task = reachable_task
        task.state.devices[task.device.name].reachable_in_last_run = False

        assert (await task.run()) is None
        assert task.state.devices[task.device.name].reachable_in_last_run is True

    async def test_run_should_update_event_to_noresponse_when_device_is_unreachable(self, unreachable_task):
        task = unreachable_task
        event = task.state.events.create_event(task.device.name, None, ReachabilityEvent)
        event.reachability = ReachabilityState.REACHABLE
        task.state.events.commit(event)

        with pytest.raises(DeviceUnreachableError):
            await task.run()
        updated_event = task.state.events[event.id]
        assert updated_event.reachability == ReachabilityState.NORESPONSE

    async def test_run_extra_job_should_update_event_to_reachable_when_device_is_reachable(self, reachable_task):
        task = reachable_task
        event = task.state.events.create_event(task.device.name, None, ReachabilityEvent)
        event.reachability = ReachabilityState.NORESPONSE
        event.lasttrans = now() - timedelta(minutes=5)
        task.state.events.commit(event)

        assert (await task._run_extra_job()) is None
        updated_event = task.state.events[event.id]
        assert updated_event.reachability == ReachabilityState.REACHABLE

    async def test_run_extra_job_should_not_update_event_when_device_is_unreachable(self, unreachable_task):
        task = unreachable_task
        event = task.state.events.create_event(task.device.name, None, ReachabilityEvent)
        event.reachability = ReachabilityState.NORESPONSE
        assert (await task._run_extra_job()) is None
        assert event.reachability == ReachabilityState.NORESPONSE

    @pytest.mark.asyncio
    async def test_when_event_is_new_it_should_set_lasttrans(self, unreachable_task):
        task = unreachable_task
        with pytest.raises(DeviceUnreachableError):
            await task.run()
        event = task.state.events.get(task.device.name, None, ReachabilityEvent)
        assert event.lasttrans

    @pytest.mark.asyncio
    async def test_when_event_transitions_from_no_response_to_reachable_it_should_update_lasttrans(
        self, reachable_task
    ):
        task = reachable_task
        initial_lasttrans = now() - timedelta(minutes=5)

        event = task.state.events.create_event(task.device.name, None, ReachabilityEvent)
        event.reachability = ReachabilityState.NORESPONSE
        event.lasttrans = initial_lasttrans
        task.state.events.commit(event)

        assert (await task.run()) is None
        updated_event = task.state.events[event.id]
        assert updated_event.reachability == ReachabilityState.REACHABLE
        assert updated_event.lasttrans > initial_lasttrans

    @pytest.mark.asyncio
    async def test_when_event_transitions_from_reachable_to_no_response_it_should_update_lasttrans(
        self, unreachable_task
    ):
        task = unreachable_task
        initial_lasttrans = now() - timedelta(minutes=5)

        event = task.state.events.create_event(task.device.name, None, ReachabilityEvent)
        event.reachability = ReachabilityState.REACHABLE
        event.lasttrans = initial_lasttrans
        task.state.events.commit(event)

        with pytest.raises(DeviceUnreachableError):
            await task.run()
        updated_event = task.state.events[event.id]
        assert updated_event.reachability == ReachabilityState.NORESPONSE
        assert updated_event.lasttrans > initial_lasttrans

    @pytest.mark.asyncio
    async def test_when_event_transitions_from_no_response_to_reachable_it_should_update_ac_down(self, reachable_task):
        task = reachable_task
        initial_ac_down = timedelta(0)
        initial_lasttrans = now() - timedelta(minutes=5)

        event = task.state.events.create_event(task.device.name, None, ReachabilityEvent)
        event.reachability = ReachabilityState.NORESPONSE
        event.ac_down = initial_ac_down
        event.lasttrans = initial_lasttrans
        task.state.events.commit(event)

        assert (await task.run()) is None
        updated_event = task.state.events[event.id]
        assert updated_event.reachability == ReachabilityState.REACHABLE
        assert updated_event.ac_down > initial_ac_down



import logging
from typing import Generator, NamedTuple

from zino.state import ZinoState
from zino.stateconverter.utils import OldState, parse_ip
from zino.statemodels import BGPAdminStatus, BGPOperState, BGPPeerSession, IPAddress

_log = logging.getLogger(__name__)
_invalid_bgp_addrs: set[str] = set()


def set_bgp_state(
    old_state: OldState,
    new_state: ZinoState,
):
    """Updates `bgp_peers` field for devices in `state` with
    BGP session data defined in `temp_sessions`. If errors occur for one peer,
    the error is logged and the peer is discarded before the next peer is processed.
    """
    for index, session in _get_bgp_sessions(old_state):
        device_name, peer_ip = index
        device = new_state.devices.get(device_name)
        device.bgp_peers[peer_ip] = session


class BGPDevicePeerIndex(NamedTuple):
    """Defines a device and a peer using the name of the device and the IP for the peer"""

    device: str
    ip: IPAddress


def _get_bgp_sessions(old_state: OldState) -> Generator[tuple[BGPDevicePeerIndex, BGPPeerSession], None, None]:
    """Parses old state for BGP sessions and maps them to the correct device and peer IP"""
    bgp_data = _group_bgp_data_by_index(old_state)
    if _invalid_bgp_addrs:
        _log.error("Ignored invalid BGP peer addresses when converting BGP sessions: %r", _invalid_bgp_addrs)

    for index, data in bgp_data.items():
        bgp_session = BGPPeerSession(
            uptime=int(data.get("::bgpPeerUpTime", 0)),
            admin_status=BGPAdminStatus(data.get("::bgpPeerAdminState")) if "::bgpPeerAdminState" in data else None,
            oper_state=BGPOperState(data.get("::bgpPeerOperState")) if "::bgpPeerOperState" in data else None,
        )
        yield index, bgp_session


def _group_bgp_data_by_index(old_state: OldState) -> dict[BGPDevicePeerIndex, dict[str, str]]:
    """Goes through the state dict and groups BGP data by device and peer IP"""
    return_dict = dict()
    for key in ("::bgpPeerAdminState", "::bgpPeerOperState", "::bgpPeerUpTime"):
        for linedata in old_state[key]:
            try:
                ip = parse_ip(linedata.identifiers[1])
            except ValueError:
                # There is a bug in zino1 statedump where invalid IPv6 addresses are dumped
                _invalid_bgp_addrs.add(linedata.identifiers[1])
                continue
            device_name = linedata.identifiers[0]
            index = BGPDevicePeerIndex(device_name, ip)
            if index not in return_dict:
                return_dict[index] = dict()
            return_dict[index][key] = linedata.value
    return return_dict



from unittest.mock import patch

import pytest

from zino.config.models import Configuration, PollDevice
from zino.state import ZinoState
from zino.tasks import run_all_tasks, run_registered_tasks
from zino.tasks.errors import DeviceUnreachableError
from zino.tasks.task import Task


class TestRunAllTasks:
    async def test_does_not_raise_error_if_device_is_unreachable(self, unreachable_task):
        assert (await run_all_tasks(unreachable_task.device, unreachable_task.state, unreachable_task.config)) is None

    async def test_when_one_task_raises_device_unreachable_it_should_not_run_further_tasks(
        self, raising_task, non_raising_task
    ):
        mock_device = PollDevice(name="mock_device", address="127.0.0.1")
        state = ZinoState()
        config = Configuration()
        with patch("zino.tasks.get_registered_tasks") as get_registered_tasks:
            get_registered_tasks.return_value = [raising_task, non_raising_task]
            await run_all_tasks(mock_device, state, config)
            assert raising_task.was_run
            assert not non_raising_task.was_run


class TestRunRegisteredTasks:
    async def test_raises_error_if_device_is_unreachable(self, unreachable_task):
        """The rest of the registered tasks are cancelled as a result of this"""
        with pytest.raises(DeviceUnreachableError):
            await run_registered_tasks(unreachable_task.device, unreachable_task.state, unreachable_task.config)


@pytest.fixture()
def raising_task():
    class RaisingTask(Task):
        was_run = False

        async def run(self):
            RaisingTask.was_run = True
            raise DeviceUnreachableError("just testing")

    return RaisingTask


@pytest.fixture()
def non_raising_task():
    class NonRaisingTask(Task):
        was_run = False

        async def run(self):
            NonRaisingTask.was_run = True
            return True

    return NonRaisingTask



import sys

import pytest

from zino.compat import StrEnum


@pytest.mark.skipif(sys.version_info >= (3, 11), reason="StrEnum compat is only valid for Python < 3.11")
def test_strenum_value_should_be_string():
    assert str(StrTestEnum.NOT_PRESENT) == "notPresent"


@pytest.mark.skipif(sys.version_info >= (3, 11), reason="StrEnum compat is only valid for Python < 3.11")
def test_strenum_should_be_able_to_instantiate_from_string():
    assert StrTestEnum("notPresent") == StrTestEnum.NOT_PRESENT


class StrTestEnum(StrEnum):
    UP = "up"
    DOWN = "down"
    NOT_PRESENT = "notPresent"



#
# PySNMP MIB module OSPF-TRAP-MIB (http://snmplabs.com/pysmi)
# ASN.1 source file:///home/mvold/mibs/ietf/OSPF-TRAP-MIB
# Produced by pysmi-1.1.12 at Thu Jul 18 13:19:33 2024
# On host agrajag platform Linux version 6.6.37 by user mvold
# Using Python version 3.11.9 (main, Apr  2 2024, 08:25:04) [GCC 13.2.0]
#
OctetString, ObjectIdentifier, Integer = mibBuilder.importSymbols("ASN1", "OctetString", "ObjectIdentifier", "Integer")
NamedValues, = mibBuilder.importSymbols("ASN1-ENUMERATION", "NamedValues")
ConstraintsIntersection, SingleValueConstraint, ValueSizeConstraint, ConstraintsUnion, ValueRangeConstraint = mibBuilder.importSymbols("ASN1-REFINEMENT", "ConstraintsIntersection", "SingleValueConstraint", "ValueSizeConstraint", "ConstraintsUnion", "ValueRangeConstraint")
ospfNbrRestartHelperAge, ospfVirtIfState, ospfRestartExitReason, ospfRestartInterval, ospfAreaNssaTranslatorState, ospfVirtNbrArea, ospfVirtIfNeighbor, ospfRouterId, ospfNbrRestartHelperStatus, ospfVirtNbrRestartHelperAge, ospfAddressLessIf, ospfNbrRestartHelperExitReason, ospfLsdbAreaId, ospfLsdbLsid, ospfVirtNbrRtrId, ospfLsdbType, ospfNbrRtrId, ospfNbrIpAddr, ospfVirtNbrState, ospfVirtNbrRestartHelperStatus, ospfVirtNbrRestartHelperExitReason, ospfLsdbRouterId, ospfIfState, ospfAreaId, ospfIfIpAddress, ospf, ospfVirtIfAreaId, ospfRestartStatus, ospfNbrState, ospfExtLsdbLimit, ospfNbrAddressLessIndex = mibBuilder.importSymbols("OSPF-MIB", "ospfNbrRestartHelperAge", "ospfVirtIfState", "ospfRestartExitReason", "ospfRestartInterval", "ospfAreaNssaTranslatorState", "ospfVirtNbrArea", "ospfVirtIfNeighbor", "ospfRouterId", "ospfNbrRestartHelperStatus", "ospfVirtNbrRestartHelperAge", "ospfAddressLessIf", "ospfNbrRestartHelperExitReason", "ospfLsdbAreaId", "ospfLsdbLsid", "ospfVirtNbrRtrId", "ospfLsdbType", "ospfNbrRtrId", "ospfNbrIpAddr", "ospfVirtNbrState", "ospfVirtNbrRestartHelperStatus", "ospfVirtNbrRestartHelperExitReason", "ospfLsdbRouterId", "ospfIfState", "ospfAreaId", "ospfIfIpAddress", "ospf", "ospfVirtIfAreaId", "ospfRestartStatus", "ospfNbrState", "ospfExtLsdbLimit", "ospfNbrAddressLessIndex")
ModuleCompliance, ObjectGroup, NotificationGroup = mibBuilder.importSymbols("SNMPv2-CONF", "ModuleCompliance", "ObjectGroup", "NotificationGroup")
Unsigned32, ObjectIdentity, Integer32, MibScalar, MibTable, MibTableRow, MibTableColumn, ModuleIdentity, MibIdentifier, iso, NotificationType, Bits, TimeTicks, Counter64, Gauge32, Counter32, IpAddress = mibBuilder.importSymbols("SNMPv2-SMI", "Unsigned32", "ObjectIdentity", "Integer32", "MibScalar", "MibTable", "MibTableRow", "MibTableColumn", "ModuleIdentity", "MibIdentifier", "iso", "NotificationType", "Bits", "TimeTicks", "Counter64", "Gauge32", "Counter32", "IpAddress")
DisplayString, TextualConvention = mibBuilder.importSymbols("SNMPv2-TC", "DisplayString", "TextualConvention")
ospfTrap = ModuleIdentity((1, 3, 6, 1, 2, 1, 14, 16))
ospfTrap.setRevisions(('2006-11-10 00:00', '1995-01-20 12:25',))
if mibBuilder.loadTexts: ospfTrap.setLastUpdated('200611100000Z')
if mibBuilder.loadTexts: ospfTrap.setOrganization('IETF OSPF Working Group')
ospfTrapControl = MibIdentifier((1, 3, 6, 1, 2, 1, 14, 16, 1))
ospfTraps = MibIdentifier((1, 3, 6, 1, 2, 1, 14, 16, 2))
ospfSetTrap = MibScalar((1, 3, 6, 1, 2, 1, 14, 16, 1, 1), OctetString().subtype(subtypeSpec=ValueSizeConstraint(4, 4)).setFixedLength(4)).setMaxAccess("readwrite")
if mibBuilder.loadTexts: ospfSetTrap.setStatus('current')
ospfConfigErrorType = MibScalar((1, 3, 6, 1, 2, 1, 14, 16, 1, 2), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13))).clone(namedValues=NamedValues(("badVersion", 1), ("areaMismatch", 2), ("unknownNbmaNbr", 3), ("unknownVirtualNbr", 4), ("authTypeMismatch", 5), ("authFailure", 6), ("netMaskMismatch", 7), ("helloIntervalMismatch", 8), ("deadIntervalMismatch", 9), ("optionMismatch", 10), ("mtuMismatch", 11), ("duplicateRouterId", 12), ("noError", 13)))).setMaxAccess("readonly")
if mibBuilder.loadTexts: ospfConfigErrorType.setStatus('current')
ospfPacketType = MibScalar((1, 3, 6, 1, 2, 1, 14, 16, 1, 3), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(1, 2, 3, 4, 5, 6))).clone(namedValues=NamedValues(("hello", 1), ("dbDescript", 2), ("lsReq", 3), ("lsUpdate", 4), ("lsAck", 5), ("nullPacket", 6)))).setMaxAccess("readonly")
if mibBuilder.loadTexts: ospfPacketType.setStatus('current')
ospfPacketSrc = MibScalar((1, 3, 6, 1, 2, 1, 14, 16, 1, 4), IpAddress()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ospfPacketSrc.setStatus('current')
ospfVirtIfStateChange = NotificationType((1, 3, 6, 1, 2, 1, 14, 16, 2, 1)).setObjects(("OSPF-MIB", "ospfRouterId"), ("OSPF-MIB", "ospfVirtIfAreaId"), ("OSPF-MIB", "ospfVirtIfNeighbor"), ("OSPF-MIB", "ospfVirtIfState"))
if mibBuilder.loadTexts: ospfVirtIfStateChange.setStatus('current')
ospfNbrStateChange = NotificationType((1, 3, 6, 1, 2, 1, 14, 16, 2, 2)).setObjects(("OSPF-MIB", "ospfRouterId"), ("OSPF-MIB", "ospfNbrIpAddr"), ("OSPF-MIB", "ospfNbrAddressLessIndex"), ("OSPF-MIB", "ospfNbrRtrId"), ("OSPF-MIB", "ospfNbrState"))
if mibBuilder.loadTexts: ospfNbrStateChange.setStatus('current')
ospfVirtNbrStateChange = NotificationType((1, 3, 6, 1, 2, 1, 14, 16, 2, 3)).setObjects(("OSPF-MIB", "ospfRouterId"), ("OSPF-MIB", "ospfVirtNbrArea"), ("OSPF-MIB", "ospfVirtNbrRtrId"), ("OSPF-MIB", "ospfVirtNbrState"))
if mibBuilder.loadTexts: ospfVirtNbrStateChange.setStatus('current')
ospfIfConfigError = NotificationType((1, 3, 6, 1, 2, 1, 14, 16, 2, 4)).setObjects(("OSPF-MIB", "ospfRouterId"), ("OSPF-MIB", "ospfIfIpAddress"), ("OSPF-MIB", "ospfAddressLessIf"), ("OSPF-TRAP-MIB", "ospfPacketSrc"), ("OSPF-TRAP-MIB", "ospfConfigErrorType"), ("OSPF-TRAP-MIB", "ospfPacketType"))
if mibBuilder.loadTexts: ospfIfConfigError.setStatus('current')
ospfVirtIfConfigError = NotificationType((1, 3, 6, 1, 2, 1, 14, 16, 2, 5)).setObjects(("OSPF-MIB", "ospfRouterId"), ("OSPF-MIB", "ospfVirtIfAreaId"), ("OSPF-MIB", "ospfVirtIfNeighbor"), ("OSPF-TRAP-MIB", "ospfConfigErrorType"), ("OSPF-TRAP-MIB", "ospfPacketType"))
if mibBuilder.loadTexts: ospfVirtIfConfigError.setStatus('current')
ospfIfAuthFailure = NotificationType((1, 3, 6, 1, 2, 1, 14, 16, 2, 6)).setObjects(("OSPF-MIB", "ospfRouterId"), ("OSPF-MIB", "ospfIfIpAddress"), ("OSPF-MIB", "ospfAddressLessIf"), ("OSPF-TRAP-MIB", "ospfPacketSrc"), ("OSPF-TRAP-MIB", "ospfConfigErrorType"), ("OSPF-TRAP-MIB", "ospfPacketType"))
if mibBuilder.loadTexts: ospfIfAuthFailure.setStatus('current')
ospfVirtIfAuthFailure = NotificationType((1, 3, 6, 1, 2, 1, 14, 16, 2, 7)).setObjects(("OSPF-MIB", "ospfRouterId"), ("OSPF-MIB", "ospfVirtIfAreaId"), ("OSPF-MIB", "ospfVirtIfNeighbor"), ("OSPF-TRAP-MIB", "ospfConfigErrorType"), ("OSPF-TRAP-MIB", "ospfPacketType"))
if mibBuilder.loadTexts: ospfVirtIfAuthFailure.setStatus('current')
ospfIfRxBadPacket = NotificationType((1, 3, 6, 1, 2, 1, 14, 16, 2, 8)).setObjects(("OSPF-MIB", "ospfRouterId"), ("OSPF-MIB", "ospfIfIpAddress"), ("OSPF-MIB", "ospfAddressLessIf"), ("OSPF-TRAP-MIB", "ospfPacketSrc"), ("OSPF-TRAP-MIB", "ospfPacketType"))
if mibBuilder.loadTexts: ospfIfRxBadPacket.setStatus('current')
ospfVirtIfRxBadPacket = NotificationType((1, 3, 6, 1, 2, 1, 14, 16, 2, 9)).setObjects(("OSPF-MIB", "ospfRouterId"), ("OSPF-MIB", "ospfVirtIfAreaId"), ("OSPF-MIB", "ospfVirtIfNeighbor"), ("OSPF-TRAP-MIB", "ospfPacketType"))
if mibBuilder.loadTexts: ospfVirtIfRxBadPacket.setStatus('current')
ospfTxRetransmit = NotificationType((1, 3, 6, 1, 2, 1, 14, 16, 2, 10)).setObjects(("OSPF-MIB", "ospfRouterId"), ("OSPF-MIB", "ospfIfIpAddress"), ("OSPF-MIB", "ospfAddressLessIf"), ("OSPF-MIB", "ospfNbrRtrId"), ("OSPF-TRAP-MIB", "ospfPacketType"), ("OSPF-MIB", "ospfLsdbType"), ("OSPF-MIB", "ospfLsdbLsid"), ("OSPF-MIB", "ospfLsdbRouterId"))
if mibBuilder.loadTexts: ospfTxRetransmit.setStatus('current')
ospfVirtIfTxRetransmit = NotificationType((1, 3, 6, 1, 2, 1, 14, 16, 2, 11)).setObjects(("OSPF-MIB", "ospfRouterId"), ("OSPF-MIB", "ospfVirtIfAreaId"), ("OSPF-MIB", "ospfVirtIfNeighbor"), ("OSPF-TRAP-MIB", "ospfPacketType"), ("OSPF-MIB", "ospfLsdbType"), ("OSPF-MIB", "ospfLsdbLsid"), ("OSPF-MIB", "ospfLsdbRouterId"))
if mibBuilder.loadTexts: ospfVirtIfTxRetransmit.setStatus('current')
ospfOriginateLsa = NotificationType((1, 3, 6, 1, 2, 1, 14, 16, 2, 12)).setObjects(("OSPF-MIB", "ospfRouterId"), ("OSPF-MIB", "ospfLsdbAreaId"), ("OSPF-MIB", "ospfLsdbType"), ("OSPF-MIB", "ospfLsdbLsid"), ("OSPF-MIB", "ospfLsdbRouterId"))
if mibBuilder.loadTexts: ospfOriginateLsa.setStatus('current')
ospfMaxAgeLsa = NotificationType((1, 3, 6, 1, 2, 1, 14, 16, 2, 13)).setObjects(("OSPF-MIB", "ospfRouterId"), ("OSPF-MIB", "ospfLsdbAreaId"), ("OSPF-MIB", "ospfLsdbType"), ("OSPF-MIB", "ospfLsdbLsid"), ("OSPF-MIB", "ospfLsdbRouterId"))
if mibBuilder.loadTexts: ospfMaxAgeLsa.setStatus('current')
ospfLsdbOverflow = NotificationType((1, 3, 6, 1, 2, 1, 14, 16, 2, 14)).setObjects(("OSPF-MIB", "ospfRouterId"), ("OSPF-MIB", "ospfExtLsdbLimit"))
if mibBuilder.loadTexts: ospfLsdbOverflow.setStatus('current')
ospfLsdbApproachingOverflow = NotificationType((1, 3, 6, 1, 2, 1, 14, 16, 2, 15)).setObjects(("OSPF-MIB", "ospfRouterId"), ("OSPF-MIB", "ospfExtLsdbLimit"))
if mibBuilder.loadTexts: ospfLsdbApproachingOverflow.setStatus('current')
ospfIfStateChange = NotificationType((1, 3, 6, 1, 2, 1, 14, 16, 2, 16)).setObjects(("OSPF-MIB", "ospfRouterId"), ("OSPF-MIB", "ospfIfIpAddress"), ("OSPF-MIB", "ospfAddressLessIf"), ("OSPF-MIB", "ospfIfState"))
if mibBuilder.loadTexts: ospfIfStateChange.setStatus('current')
ospfNssaTranslatorStatusChange = NotificationType((1, 3, 6, 1, 2, 1, 14, 16, 2, 17)).setObjects(("OSPF-MIB", "ospfRouterId"), ("OSPF-MIB", "ospfAreaId"), ("OSPF-MIB", "ospfAreaNssaTranslatorState"))
if mibBuilder.loadTexts: ospfNssaTranslatorStatusChange.setStatus('current')
ospfRestartStatusChange = NotificationType((1, 3, 6, 1, 2, 1, 14, 16, 2, 18)).setObjects(("OSPF-MIB", "ospfRouterId"), ("OSPF-MIB", "ospfRestartStatus"), ("OSPF-MIB", "ospfRestartInterval"), ("OSPF-MIB", "ospfRestartExitReason"))
if mibBuilder.loadTexts: ospfRestartStatusChange.setStatus('current')
ospfNbrRestartHelperStatusChange = NotificationType((1, 3, 6, 1, 2, 1, 14, 16, 2, 19)).setObjects(("OSPF-MIB", "ospfRouterId"), ("OSPF-MIB", "ospfNbrIpAddr"), ("OSPF-MIB", "ospfNbrAddressLessIndex"), ("OSPF-MIB", "ospfNbrRtrId"), ("OSPF-MIB", "ospfNbrRestartHelperStatus"), ("OSPF-MIB", "ospfNbrRestartHelperAge"), ("OSPF-MIB", "ospfNbrRestartHelperExitReason"))
if mibBuilder.loadTexts: ospfNbrRestartHelperStatusChange.setStatus('current')
ospfVirtNbrRestartHelperStatusChange = NotificationType((1, 3, 6, 1, 2, 1, 14, 16, 2, 20)).setObjects(("OSPF-MIB", "ospfRouterId"), ("OSPF-MIB", "ospfVirtNbrArea"), ("OSPF-MIB", "ospfVirtNbrRtrId"), ("OSPF-MIB", "ospfVirtNbrRestartHelperStatus"), ("OSPF-MIB", "ospfVirtNbrRestartHelperAge"), ("OSPF-MIB", "ospfVirtNbrRestartHelperExitReason"))
if mibBuilder.loadTexts: ospfVirtNbrRestartHelperStatusChange.setStatus('current')
ospfTrapConformance = MibIdentifier((1, 3, 6, 1, 2, 1, 14, 16, 3))
ospfTrapGroups = MibIdentifier((1, 3, 6, 1, 2, 1, 14, 16, 3, 1))
ospfTrapCompliances = MibIdentifier((1, 3, 6, 1, 2, 1, 14, 16, 3, 2))
ospfTrapCompliance = ModuleCompliance((1, 3, 6, 1, 2, 1, 14, 16, 3, 2, 1)).setObjects(("OSPF-TRAP-MIB", "ospfTrapControlGroup"), ("OSPF-TRAP-MIB", "ospfTrapControlGroup"))

if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ospfTrapCompliance = ospfTrapCompliance.setStatus('obsolete')
ospfTrapCompliance2 = ModuleCompliance((1, 3, 6, 1, 2, 1, 14, 16, 3, 2, 2)).setObjects(("OSPF-TRAP-MIB", "ospfTrapControlGroup"), ("OSPF-TRAP-MIB", "ospfTrapEventGroup"))

if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ospfTrapCompliance2 = ospfTrapCompliance2.setStatus('current')
ospfTrapControlGroup = ObjectGroup((1, 3, 6, 1, 2, 1, 14, 16, 3, 1, 1)).setObjects(("OSPF-TRAP-MIB", "ospfSetTrap"), ("OSPF-TRAP-MIB", "ospfConfigErrorType"), ("OSPF-TRAP-MIB", "ospfPacketType"), ("OSPF-TRAP-MIB", "ospfPacketSrc"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ospfTrapControlGroup = ospfTrapControlGroup.setStatus('current')
ospfTrapEventGroup = NotificationGroup((1, 3, 6, 1, 2, 1, 14, 16, 3, 1, 2)).setObjects(("OSPF-TRAP-MIB", "ospfVirtIfStateChange"), ("OSPF-TRAP-MIB", "ospfNbrStateChange"), ("OSPF-TRAP-MIB", "ospfVirtNbrStateChange"), ("OSPF-TRAP-MIB", "ospfIfConfigError"), ("OSPF-TRAP-MIB", "ospfVirtIfConfigError"), ("OSPF-TRAP-MIB", "ospfIfAuthFailure"), ("OSPF-TRAP-MIB", "ospfVirtIfAuthFailure"), ("OSPF-TRAP-MIB", "ospfIfRxBadPacket"), ("OSPF-TRAP-MIB", "ospfVirtIfRxBadPacket"), ("OSPF-TRAP-MIB", "ospfTxRetransmit"), ("OSPF-TRAP-MIB", "ospfVirtIfTxRetransmit"), ("OSPF-TRAP-MIB", "ospfOriginateLsa"), ("OSPF-TRAP-MIB", "ospfMaxAgeLsa"), ("OSPF-TRAP-MIB", "ospfLsdbOverflow"), ("OSPF-TRAP-MIB", "ospfLsdbApproachingOverflow"), ("OSPF-TRAP-MIB", "ospfIfStateChange"), ("OSPF-TRAP-MIB", "ospfNssaTranslatorStatusChange"), ("OSPF-TRAP-MIB", "ospfRestartStatusChange"), ("OSPF-TRAP-MIB", "ospfNbrRestartHelperStatusChange"), ("OSPF-TRAP-MIB", "ospfVirtNbrRestartHelperStatusChange"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ospfTrapEventGroup = ospfTrapEventGroup.setStatus('current')
mibBuilder.exportSymbols("OSPF-TRAP-MIB", ospfTrapGroups=ospfTrapGroups, ospfSetTrap=ospfSetTrap, ospfNbrStateChange=ospfNbrStateChange, ospfTrapCompliances=ospfTrapCompliances, ospfVirtIfTxRetransmit=ospfVirtIfTxRetransmit, ospfTrapCompliance=ospfTrapCompliance, ospfTraps=ospfTraps, ospfVirtNbrRestartHelperStatusChange=ospfVirtNbrRestartHelperStatusChange, ospfNbrRestartHelperStatusChange=ospfNbrRestartHelperStatusChange, ospfTrap=ospfTrap, ospfVirtNbrStateChange=ospfVirtNbrStateChange, ospfTrapControl=ospfTrapControl, ospfTrapConformance=ospfTrapConformance, ospfIfConfigError=ospfIfConfigError, ospfConfigErrorType=ospfConfigErrorType, ospfTxRetransmit=ospfTxRetransmit, ospfVirtIfStateChange=ospfVirtIfStateChange, ospfRestartStatusChange=ospfRestartStatusChange, ospfVirtIfAuthFailure=ospfVirtIfAuthFailure, ospfLsdbOverflow=ospfLsdbOverflow, ospfTrapCompliance2=ospfTrapCompliance2, ospfIfRxBadPacket=ospfIfRxBadPacket, ospfVirtIfRxBadPacket=ospfVirtIfRxBadPacket, ospfPacketSrc=ospfPacketSrc, ospfNssaTranslatorStatusChange=ospfNssaTranslatorStatusChange, ospfIfAuthFailure=ospfIfAuthFailure, ospfMaxAgeLsa=ospfMaxAgeLsa, ospfTrapControlGroup=ospfTrapControlGroup, ospfPacketType=ospfPacketType, ospfIfStateChange=ospfIfStateChange, PYSNMP_MODULE_ID=ospfTrap, ospfLsdbApproachingOverflow=ospfLsdbApproachingOverflow, ospfOriginateLsa=ospfOriginateLsa, ospfTrapEventGroup=ospfTrapEventGroup, ospfVirtIfConfigError=ospfVirtIfConfigError)



import logging

from zino.state import ZinoState
from zino.stateconverter.linedata import LineData
from zino.stateconverter.utils import OldState
from zino.statemodels import InterfaceState, Port

_log = logging.getLogger(__name__)


def set_port_state(old_state: OldState, new_state: ZinoState):
    for linedata in old_state["::portState"]:
        _set_interface_state(linedata, new_state)
    for linedata in old_state["::portToIfDescr"]:
        _set_port_to_if_descr(linedata, new_state)
    for linedata in old_state["::portToLocIfDescr"]:
        _set_port_to_loc_if_descr(linedata, new_state)


def _set_interface_state(linedata: LineData, state: ZinoState):
    device = state.devices.get(linedata.identifiers[0])
    ifindex = int(linedata.identifiers[1])
    if ifindex not in device.ports:
        device.ports[ifindex] = Port(ifindex=ifindex)
    device.ports[ifindex].state = InterfaceState(linedata.value)


def _set_port_to_if_descr(linedata: LineData, state: ZinoState):
    device = state.devices.get(linedata.identifiers[0])
    ifindex = int(linedata.identifiers[1])
    if ifindex not in device.ports:
        device.ports[ifindex] = Port(ifindex=ifindex)
    device.ports[ifindex].ifdescr = linedata.value


def _set_port_to_loc_if_descr(linedata: LineData, state: ZinoState):
    """Sets ifalias value"""
    device = state.devices.get(linedata.identifiers[0])
    ifindex = int(linedata.identifiers[1])
    if ifindex not in device.ports:
        device.ports[ifindex] = Port(ifindex=ifindex)
    device.ports[ifindex].ifalias = linedata.value



"""SNMP agent for Zino that responds to uptime queries.

This module implements an SNMP agent that listens on a configurable port
and responds to queries for ZINO-MIB::zinoUpTime, which is used by clients
for failover detection.

The agent is built on PySNMP, since the netsnmpy library does not yet provide
support for setting up SNMP agents - and performance optimizations for many
parallel conversations are not needed here.
"""

import asyncio
import logging
import os
import time
from ipaddress import IPv6Address, ip_address
from typing import Optional

from pysnmp.carrier.asyncio.dgram import udp, udp6
from pysnmp.carrier.asyncio.dispatch import AsyncioDispatcher
from pysnmp.entity import config, engine
from pysnmp.entity.rfc3413 import cmdrsp, context
from pysnmp.smi import builder

_log = logging.getLogger(__name__)


class ZinoSnmpAgent:
    """SNMP agent that responds to queries for Zino uptime.

    This agent is used by clients like EMT for failover detection.
    It responds to SNMP GET requests for the ZINO-MIB::zinoUpTime object.
    """

    def __init__(
        self,
        listen_address: str = "0.0.0.0",
        listen_port: int = 8000,
        community: Optional[str] = None,
        start_time: Optional[float] = None,
    ):
        """Initialize the SNMP agent.

        :param listen_address: IP address to bind to (default: "0.0.0.0")
        :param listen_port: Port to listen on (default: 8000)
        :param community: SNMP community string (default: None, accepts any)
        :param start_time: Server start timestamp (default: current time)
        """
        self.listen_address = listen_address
        self.listen_port = listen_port
        self.community = community or "public"
        self.start_time = start_time or time.time()
        self.snmp_engine = None
        self.transport_dispatcher = None
        self._running = False

    def _setup_engine(self):
        """Sets up the SNMP engine with configuration."""
        # Create SNMP engine
        self.snmp_engine = engine.SnmpEngine()

        # Configure transport (UDP/IPv4 and IPv6)
        self.transport_dispatcher = AsyncioDispatcher()
        self.snmp_engine.registerTransportDispatcher(self.transport_dispatcher)

        # Determine if we need IPv4 or IPv6 transport
        try:
            addr = ip_address(self.listen_address)
            if isinstance(addr, IPv6Address):
                transport = udp6.Udp6AsyncioTransport()
                endpoint = (self.listen_address, self.listen_port)
            else:
                transport = udp.UdpAsyncioTransport()
                endpoint = (self.listen_address, self.listen_port)
        except ValueError:
            # If address is a hostname, default to IPv4
            transport = udp.UdpAsyncioTransport()
            endpoint = (self.listen_address, self.listen_port)

        # Open the transport - this actually binds to the socket
        transport = transport.openServerMode(endpoint)

        # Add transport to the engine
        config.addTransport(self.snmp_engine, udp.domainName, transport)

        # Configure SNMPv1 and SNMPv2c
        config.addV1System(self.snmp_engine, "zino-agent", self.community)

        # Configure vacm to allow access
        config.addVacmUser(self.snmp_engine, 2, "zino-agent", "noAuthNoPriv", (), (), ())  # SNMPv2c
        config.addVacmUser(self.snmp_engine, 1, "zino-agent", "noAuthNoPriv", (), (), ())  # SNMPv1

        # Get MIB controller and builder
        mib_instrum = self.snmp_engine.msgAndPduDsp.mibInstrumController
        mib_builder = mib_instrum.mibBuilder

        # Add the mibdumps directory to the MIB search path
        mibdumps_dir = os.path.join(os.path.dirname(__file__), "mibdumps")
        mib_builder.addMibSources(builder.DirMibSource(mibdumps_dir))

        # Load the required MIB modules
        mib_builder.loadModules("SNMPv2-MIB", "UNINETT-SMI", "ZINO-MIB")

        # Register the zinoUpTime scalar with dynamic value
        self._register_zino_uptime(mib_instrum)

        # Register SNMP command responders for GET and GETNEXT
        snmp_context = context.SnmpContext(self.snmp_engine)
        cmdrsp.GetCommandResponder(self.snmp_engine, snmp_context)
        cmdrsp.NextCommandResponder(self.snmp_engine, snmp_context)

        _log.debug("SNMP agent configured on %s:%d", self.listen_address, self.listen_port)

    def _register_zino_uptime(self, mib_instrum):
        """Registers zinoUpTime MIB object with dynamic uptime calculation."""
        mib_builder = mib_instrum.mibBuilder

        # Import zinoUpTime from loaded MIB and MibScalarInstance
        (zino_uptime,) = mib_builder.importSymbols("ZINO-MIB", "zinoUpTime")
        (MibScalarInstance,) = mib_builder.importSymbols("SNMPv2-SMI", "MibScalarInstance")

        # Create custom instance that returns current uptime
        class UptimeInstance(MibScalarInstance):
            def readGet(self, name, val, idx, acInfo):
                return name, self.syntax.clone(self._get_uptime())

        # Register the dynamic instance
        instance = UptimeInstance(zino_uptime.getName(), (0,), zino_uptime.getSyntax())
        instance._get_uptime = self._get_uptime  # Bind the uptime method

        mib_builder.exportSymbols("__UPTIME__", zinoUpTimeInstance=instance)
        mib_builder.loadModules("__UPTIME__")

    def _get_uptime(self) -> int:
        """Calculates uptime in seconds since instance was created."""
        return int(time.time() - self.start_time)

    async def open(self):
        """Opens the SNMP agent port and start serving SNMP requests."""
        if self._running:
            _log.warning("SNMP agent is already running")
            return

        try:
            self._setup_engine()
            self._running = True
            _log.info("SNMP agent started on %s:%d", self.listen_address, self.listen_port)
        except Exception as error:
            _log.error("Failed to start SNMP agent: %s", error)
            self._running = False
            raise

    def close(self):
        """Closes the port and stop the SNMP agent."""
        if not self._running:
            return

        _log.info("Stopping SNMP agent")
        if self.transport_dispatcher:
            self.transport_dispatcher.closeDispatcher()
        if self.snmp_engine:
            self.snmp_engine.transportDispatcher.closeDispatcher()
        self._running = False


def main():
    """Run the SNMP agent standalone for testing."""
    import argparse

    parser = argparse.ArgumentParser(description="Zino SNMP Agent")
    parser.add_argument("--host", default="0.0.0.0", help="Bind address (default: 0.0.0.0)")
    parser.add_argument("--port", type=int, default=8000, help="Port to listen on (default: 8000)")
    parser.add_argument("--community", default=None, help="SNMP community string (default: accept any)")
    parser.add_argument("--debug", action="store_true", help="Enable debug logging")

    args = parser.parse_args()

    # Set debug logging if requested
    if args.debug:
        logging.getLogger().setLevel(logging.DEBUG)
        _log.setLevel(logging.DEBUG)

    agent = ZinoSnmpAgent(listen_address=args.host, listen_port=args.port, community=args.community)

    print(f"Starting SNMP agent on {args.host}:{args.port}")
    print(f"Community: {args.community if args.community else 'any'}")
    print("OID: ZINO-MIB::zinoUpTime")
    print("Press Ctrl+C to stop...")

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    loop.run_until_complete(agent.open())

    try:
        loop.run_forever()
    except KeyboardInterrupt:
        print("\nStopping agent...")
        agent.close()
    finally:
        loop.close()


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    main()



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional

from zino.config.models import Configuration, PollDevice
from zino.snmp import get_snmp_session

if TYPE_CHECKING:
    from zino.state import ZinoState

from zino.statemodels import DeviceState


class Task(ABC):
    def __init__(self, device: PollDevice, state: ZinoState, config: Optional[Configuration] = None):
        self.device = device
        self.state = state
        self.config = config
        self.snmp = get_snmp_session(device=device)

    @abstractmethod
    async def run(self):
        """Runs job asynchronously"""

    async def _get_uptime(self) -> int:
        """Polls and returns the device sysuptime value"""
        response = await self.snmp.get("SNMPv2-MIB", "sysUpTime", 0)
        uptime = response.value
        return uptime

    @property
    def device_state(self) -> DeviceState:
        return self.state.devices.get(self.device.name)



import asyncio
import logging
import os
import stat
from functools import wraps
from time import time
from typing import Optional, Union

import aiodns


async def reverse_dns(ip: str) -> Optional[str]:
    """Returns hostname for given IP address or None if reverse DNS lookup fails"""
    loop = asyncio.get_event_loop()
    resolver = aiodns.DNSResolver(loop=loop)
    try:
        response = await resolver.gethostbyaddr(ip)
        return response.name
    except aiodns.error.DNSError:
        return None


def log_time_spent(logger: Union[logging.Logger, str] = __name__, level: int = logging.DEBUG):
    """Decorator that logs the time taken for a function to execute.  Not suitable for use with async functions"""
    if isinstance(logger, str):
        logger = logging.getLogger(logger)

    def actual_decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            start = time()
            try:
                result = func(*args, **kwargs)
            finally:
                end = time()
                logger.log(level, "%s took %s seconds", func.__name__, end - start)
            return result

        return wrapper

    return actual_decorator


def file_is_world_readable(file: str) -> bool:
    """Returns a boolean value indicating if a file is readable by other users than its owner"""
    st_mode = getattr(os.stat(path=file), "st_mode", None)

    return bool(st_mode & stat.S_IROTH)



"""Test suite for the JobTracker component."""

import asyncio
import signal
from datetime import datetime, timedelta
from unittest.mock import Mock, patch

import pytest
from apscheduler.events import (
    EVENT_JOB_ERROR,
    EVENT_JOB_EXECUTED,
    EVENT_JOB_MAX_INSTANCES,
    EVENT_JOB_MISSED,
    EVENT_JOB_SUBMITTED,
    JobExecutionEvent,
    JobSubmissionEvent,
)

from zino.job_tracker import JobTracker, get_job_tracker


class TestJobTracker:
    """Test the JobTracker class functionality."""

    @pytest.fixture
    def mock_scheduler(self):
        """Create a mock scheduler."""
        scheduler = Mock()
        scheduler.get_job = Mock(return_value=None)
        return scheduler

    def test_setup_signal_handler_should_register_sigusr1_with_event_loop(self):
        """Test setting up signal handler for SIGUSR1."""
        job_tracker = JobTracker()
        mock_loop = Mock(spec=asyncio.AbstractEventLoop)
        job_tracker.setup_signal_handler(mock_loop)

        assert job_tracker.loop == mock_loop
        mock_loop.add_signal_handler.assert_called_once_with(signal.SIGUSR1, job_tracker._handle_sigusr1)

    def test_when_no_scheduler_is_set_get_job_name_should_return_none(self):
        """Test _get_job_name when no scheduler is set."""
        job_tracker = JobTracker()
        assert job_tracker._get_job_name("job123") is None

    def test_when_scheduler_has_job_get_job_name_should_return_name(self, mock_scheduler):
        """Test _get_job_name when scheduler is available."""
        job_tracker = JobTracker()
        mock_job = Mock()
        mock_job.name = "TestJobName"
        mock_scheduler.get_job.return_value = mock_job

        job_tracker.scheduler = mock_scheduler
        result = job_tracker._get_job_name("job123")

        assert result == "TestJobName"
        mock_scheduler.get_job.assert_called_once_with("job123")

    def test_when_scheduler_raises_exception_get_job_name_should_return_none(self, mock_scheduler):
        """Test _get_job_name when scheduler raises an exception."""
        job_tracker = JobTracker()
        mock_scheduler.get_job.side_effect = Exception("Scheduler error")

        job_tracker.scheduler = mock_scheduler
        result = job_tracker._get_job_name("job123")

        assert result is None

    def test_format_duration_should_format_seconds_without_days(self):
        """Test formatting duration less than a day."""
        job_tracker = JobTracker()
        assert job_tracker._format_duration(0) == "00:00:00.000"
        assert job_tracker._format_duration(1.5) == "00:00:01.500"
        assert job_tracker._format_duration(61.123) == "00:01:01.123"
        assert job_tracker._format_duration(3661.456) == "01:01:01.456"
        assert job_tracker._format_duration(3600 * 23 + 3599.999) == "23:59:59.999"

    def test_format_duration_should_include_days_when_over_24_hours(self):
        """Test formatting duration with days."""
        job_tracker = JobTracker()
        assert job_tracker._format_duration(86400) == "1d 00:00:00.000"
        assert job_tracker._format_duration(90061.5) == "1d 01:01:01.500"
        assert job_tracker._format_duration(86400 * 10 + 7200 + 180 + 1.234) == "10d 02:03:01.234"

    @patch("zino.job_tracker.datetime")
    def test_when_no_jobs_executing_sigusr1_should_log_empty_message(self, mock_datetime, caplog):
        """Test SIGUSR1 handler when no jobs are running."""
        caplog.set_level("INFO")
        job_tracker = JobTracker()
        mock_datetime.now.return_value = datetime(2024, 1, 1, 12, 0, 0)

        job_tracker._handle_sigusr1()

        # Check that it logs "No jobs currently running"
        assert "No jobs currently running" in caplog.text

    @patch("zino.job_tracker.datetime")
    def test_when_jobs_are_executing_sigusr1_should_log_jobs(self, mock_datetime, caplog):
        """Test SIGUSR1 handler with running jobs."""
        caplog.set_level("INFO")
        job_tracker = JobTracker()
        # Set up mock time
        current_time = datetime(2024, 1, 1, 12, 0, 0)
        mock_datetime.now.return_value = current_time

        # Add some running jobs with different characteristics
        job_tracker.running_jobs = {
            "job1": {
                "start_time": current_time - timedelta(seconds=120),
                "job_name": "TestJob1",
            },
            "job2": {
                "start_time": current_time - timedelta(seconds=60),
                "job_name": None,
            },
            "job3": {
                "start_time": current_time - timedelta(seconds=300),
                "job_name": "job3",  # Same as ID, should not show in parentheses
            },
        }

        job_tracker._handle_sigusr1()

        # Verify the summary line
        assert "Total running jobs: 3" in caplog.text

        # Verify table header is logged
        assert "Job ID" in caplog.text
        assert "Duration" in caplog.text
        assert "Started" in caplog.text

        # Verify each job appears in the output
        assert "job1 (TestJob1)" in caplog.text, "job1 with name TestJob1 should appear"
        assert "job2" in caplog.text, "job2 should appear"
        assert "job3" in caplog.text, "job3 should appear"
        # job3 shouldn't show name in parentheses since it matches the ID
        assert "job3 (job3)" not in caplog.text, "job3 should not show redundant name in parentheses"

    @patch("zino.job_tracker.datetime")
    def test_sigusr1_should_format_job_table_sorted_by_duration(self, mock_datetime, caplog):
        """Test the formatted output of SIGUSR1 handler is sorted by duration."""
        caplog.set_level("INFO")
        job_tracker = JobTracker()
        current_time = datetime(2024, 1, 1, 12, 0, 0)
        mock_datetime.now.return_value = current_time

        # Add jobs with varying durations
        job_tracker.running_jobs = {
            "job1": {
                "start_time": current_time - timedelta(days=1, hours=2, minutes=30, seconds=45.123),
                "job_name": "LongRunningTask",
            },
            "job2": {
                "start_time": current_time - timedelta(minutes=5, seconds=30.5),
                "job_name": "QuickTask",
            },
            "job3": {
                "start_time": current_time - timedelta(hours=1),
                "job_name": None,
            },
        }

        job_tracker._handle_sigusr1()

        # Verify the summary line
        assert "Total running jobs: 3" in caplog.text

        # Verify the output includes all jobs and is properly formatted
        assert "Job ID" in caplog.text
        assert "Duration" in caplog.text
        assert "Started" in caplog.text
        assert "LongRunningTask" in caplog.text
        assert "QuickTask" in caplog.text

        # Verify jobs are sorted by duration (longest first)
        # Find the positions of each job in the output
        longrunning_pos = caplog.text.find("LongRunningTask")
        job3_pos = caplog.text.find("job3")
        quicktask_pos = caplog.text.find("QuickTask")

        # All jobs should be present
        assert longrunning_pos != -1, "LongRunningTask not found in output"
        assert job3_pos != -1, "job3 not found in output"
        assert quicktask_pos != -1, "QuickTask not found in output"

        # Verify order: LongRunningTask (1d 2h) should come before job3 (1h) which should come before QuickTask (5m)
        assert longrunning_pos < job3_pos, "LongRunningTask should appear before job3 (sorted by duration)"
        assert job3_pos < quicktask_pos, "job3 should appear before QuickTask (sorted by duration)"

    def test_on_job_submitted_should_track_new_job(self, mock_scheduler):
        """Test tracking when a new job starts."""
        job_tracker = JobTracker()
        job_tracker.scheduler = mock_scheduler

        # Create mock job with name attribute properly set
        mock_job = Mock()
        mock_job.name = "TestJob"
        mock_scheduler.get_job.return_value = mock_job

        event = Mock(spec=JobSubmissionEvent)
        event.job_id = "job123"
        event.scheduled_run_times = [datetime(2024, 1, 1, 10, 0, 0)]
        event.jobstore = "default"

        job_tracker.on_job_submitted(event)

        assert "job123" in job_tracker.running_jobs
        assert job_tracker.running_jobs["job123"]["scheduled_run_time"] == datetime(2024, 1, 1, 10, 0, 0)
        assert job_tracker.running_jobs["job123"]["jobstore"] == "default"
        assert job_tracker.running_jobs["job123"]["job_name"] == "TestJob"
        assert "start_time" in job_tracker.running_jobs["job123"]

    def test_when_job_already_running_on_job_submitted_should_not_update(self):
        """Test when a job is submitted but already in running list."""
        job_tracker = JobTracker()
        job_tracker.running_jobs["job123"] = {"start_time": datetime.now()}

        event = Mock(spec=JobSubmissionEvent)
        event.job_id = "job123"
        event.scheduled_run_times = [datetime.now()]
        event.jobstore = "default"

        original_info = job_tracker.running_jobs["job123"]
        job_tracker.on_job_submitted(event)

        # Should not update the existing entry
        assert job_tracker.running_jobs["job123"] == original_info

    @patch("zino.job_tracker.datetime")
    def test_when_no_scheduled_time_on_job_submitted_should_use_current_time(self, mock_datetime):
        """Test when job has no scheduled run times."""
        job_tracker = JobTracker()
        current_time = datetime(2024, 1, 1, 12, 0, 0)
        mock_datetime.now.return_value = current_time

        event = Mock(spec=JobSubmissionEvent)
        event.job_id = "job123"
        event.scheduled_run_times = []
        event.jobstore = "default"

        job_tracker.on_job_submitted(event)

        assert "job123" in job_tracker.running_jobs
        assert job_tracker.running_jobs["job123"]["scheduled_run_time"] == current_time

    @patch("zino.job_tracker.datetime")
    def test_on_job_executed_should_remove_job_from_running(self, mock_datetime):
        """Test tracking when a job completes successfully."""
        job_tracker = JobTracker()
        start_time = datetime(2024, 1, 1, 12, 0, 0)
        end_time = datetime(2024, 1, 1, 12, 0, 10)
        mock_datetime.now.side_effect = [end_time]

        job_tracker.running_jobs["job123"] = {"start_time": start_time}

        event = Mock(spec=JobExecutionEvent)
        event.job_id = "job123"

        job_tracker.on_job_executed(event)

        assert "job123" not in job_tracker.running_jobs

    def test_when_job_not_tracked_on_job_executed_should_not_raise(self):
        """Test job execution event for job not in running list."""
        job_tracker = JobTracker()
        event = Mock(spec=JobExecutionEvent)
        event.job_id = "job123"

        # Should not raise an exception
        job_tracker.on_job_executed(event)
        assert "job123" not in job_tracker.running_jobs

    @patch("zino.job_tracker.datetime")
    def test_on_job_error_should_remove_job_from_running(self, mock_datetime):
        """Test tracking when a job fails with error."""
        job_tracker = JobTracker()
        start_time = datetime(2024, 1, 1, 12, 0, 0)
        end_time = datetime(2024, 1, 1, 12, 0, 10)
        mock_datetime.now.side_effect = [end_time]

        job_tracker.running_jobs["job123"] = {"start_time": start_time}

        event = Mock(spec=JobExecutionEvent)
        event.job_id = "job123"
        event.exception = Exception("Test error")

        job_tracker.on_job_error(event)

        assert "job123" not in job_tracker.running_jobs

    def test_when_job_not_tracked_on_job_error_should_not_raise(self):
        """Test job error event for job not in running list."""
        job_tracker = JobTracker()
        event = Mock(spec=JobExecutionEvent)
        event.job_id = "job123"
        event.exception = Exception("Test error")

        # Should not raise an exception
        job_tracker.on_job_error(event)
        assert "job123" not in job_tracker.running_jobs

    def test_when_missed_job_in_running_list_it_should_be_removed(self):
        """Test when a missed job is unexpectedly in running list."""
        job_tracker = JobTracker()
        job_tracker.running_jobs["job123"] = {"start_time": datetime.now()}

        event = Mock(spec=JobExecutionEvent)
        event.job_id = "job123"

        job_tracker.on_job_missed(event)

        assert "job123" not in job_tracker.running_jobs

    def test_when_job_missed_it_should_not_be_in_running_list(self):
        """Test normal case of missed job not in running list."""
        job_tracker = JobTracker()
        event = Mock(spec=JobExecutionEvent)
        event.job_id = "job123"

        job_tracker.on_job_missed(event)
        assert "job123" not in job_tracker.running_jobs

    @patch("zino.job_tracker.datetime")
    def test_when_max_instances_reached_running_job_should_remain_tracked(self, mock_datetime):
        """Test max_instances event when job is running."""
        job_tracker = JobTracker()
        start_time = datetime(2024, 1, 1, 12, 0, 0)
        current_time = datetime(2024, 1, 1, 12, 5, 0)
        mock_datetime.now.return_value = current_time

        job_tracker.running_jobs["job123"] = {"start_time": start_time}

        event = Mock(spec=JobExecutionEvent)
        event.job_id = "job123"

        job_tracker.on_job_max_instances(event)

        # Job should still be in running list
        assert "job123" in job_tracker.running_jobs
        assert job_tracker.running_jobs["job123"]["start_time"] == start_time

    def test_when_max_instances_reached_for_untracked_job_it_should_stay_untracked(self):
        """Test max_instances event when job is not in running list."""
        job_tracker = JobTracker()
        event = Mock(spec=JobExecutionEvent)
        event.job_id = "job123"

        job_tracker.on_job_max_instances(event)
        assert "job123" not in job_tracker.running_jobs

    def test_register_with_scheduler_should_add_all_event_listeners(self, mock_scheduler):
        """Test registering tracker with scheduler."""
        job_tracker = JobTracker()
        job_tracker.register_with_scheduler(mock_scheduler)

        assert job_tracker.scheduler == mock_scheduler
        assert mock_scheduler.add_listener.call_count == 5

        # Check each listener registration
        calls = mock_scheduler.add_listener.call_args_list
        assert calls[0][0][0] == job_tracker.on_job_submitted
        assert calls[0][0][1] == EVENT_JOB_SUBMITTED
        assert calls[1][0][0] == job_tracker.on_job_executed
        assert calls[1][0][1] == EVENT_JOB_EXECUTED
        assert calls[2][0][0] == job_tracker.on_job_error
        assert calls[2][0][1] == EVENT_JOB_ERROR
        assert calls[3][0][0] == job_tracker.on_job_missed
        assert calls[3][0][1] == EVENT_JOB_MISSED
        assert calls[4][0][0] == job_tracker.on_job_max_instances
        assert calls[4][0][1] == EVENT_JOB_MAX_INSTANCES


class TestJobTrackerSingleton:
    """Test the singleton pattern for JobTracker."""

    def test_get_job_tracker_should_create_singleton_instance(self):
        """Test that get_job_tracker creates a singleton instance."""
        # Reset the global variable first
        import zino.job_tracker

        zino.job_tracker._job_tracker = None

        tracker1 = get_job_tracker()
        tracker2 = get_job_tracker()

        assert tracker1 is tracker2
        assert isinstance(tracker1, JobTracker)

    def test_get_job_tracker_should_reuse_existing_instance(self):
        """Test that get_job_tracker reuses existing instance."""
        import zino.job_tracker

        # Set up an existing tracker
        existing_tracker = JobTracker()
        zino.job_tracker._job_tracker = existing_tracker

        result = get_job_tracker()

        assert result is existing_tracker



import datetime
from typing import Iterator
from unittest.mock import patch

import pytest

from zino.statemodels import (
    EventState,
    MatchType,
    PortStateEvent,
    PortStateMaintenance,
    ReachabilityEvent,
)


class TestStart:
    def test_should_set_state_of_matching_events_to_ignored(self, state, portstate_event, matching_portstate_pm):
        assert portstate_event.state is EventState.OPEN
        matching_portstate_pm.start(state)
        assert (
            state.events.get(portstate_event.router, portstate_event.subindex, PortStateEvent).state
            is EventState.IGNORED
        )

    def test_should_add_started_events_to_event_ids_list(self, state, portstate_event, matching_portstate_pm):
        assert portstate_event.id not in matching_portstate_pm.event_ids
        matching_portstate_pm.start(state)
        assert portstate_event.id in matching_portstate_pm.event_ids

    def test_should_not_change_state_of_events_that_are_not_portstate_events(
        self, state, reachability_event, matching_portstate_pm
    ):
        assert reachability_event.state is EventState.OPEN
        matching_portstate_pm.start(state)
        assert (
            state.events.get(reachability_event.router, reachability_event.subindex, ReachabilityEvent).state
            is EventState.OPEN
        )

    def test_should_create_matching_event_with_ignored_state_if_it_does_not_exist(
        self, state, device, port, matching_portstate_pm
    ):
        assert not state.events.get(device.name, port.ifindex, PortStateEvent)
        matching_portstate_pm.start(state)
        created_event = state.events.get(device.name, port.ifindex, PortStateEvent)
        assert created_event
        assert created_event.state is EventState.IGNORED

    def test_should_not_change_state_of_non_matching_events(self, state, portstate_event, nonmatching_portstate_pm):
        assert portstate_event.state is EventState.OPEN
        nonmatching_portstate_pm.start(state)
        assert (
            state.events.get(portstate_event.router, portstate_event.subindex, PortStateEvent).state is EventState.OPEN
        )


class TestMatchesEvent:
    def test_should_return_true_for_matching_event(self, state, portstate_event, matching_portstate_pm):
        assert matching_portstate_pm.matches_event(portstate_event, state)

    def test_should_return_false_for_events_that_are_not_portstate_events(
        self, state, reachability_event, matching_portstate_pm
    ):
        assert not matching_portstate_pm.matches_event(reachability_event, state)

    def test_should_return_false_for_non_matching_events(self, state, reachability_event, nonmatching_portstate_pm):
        assert not nonmatching_portstate_pm.matches_event(reachability_event, state)


class TestMatchesPortstate:
    @pytest.mark.parametrize("portstate_pm", [MatchType.REGEXP, MatchType.STR, MatchType.INTF_REGEXP], indirect=True)
    def test_should_return_false_for_non_matching_port(self, portstate_pm, device, port):
        port.ifalias = "wrong"
        port.ifdescr = "wrong"
        assert not portstate_pm.matches_portstate(device, port)

    @pytest.mark.parametrize("portstate_pm", [MatchType.REGEXP, MatchType.STR], indirect=True)
    def test_regexp_and_str_should_return_true_for_matching_port(self, portstate_pm, device, port):
        assert portstate_pm.matches_portstate(device, port)

    @pytest.mark.parametrize("portstate_pm", [MatchType.REGEXP], indirect=True)
    def test_regexp_should_return_true_for_matching_port_not_at_start_of_regexp_expression(
        self, portstate_pm, device, port
    ):
        port.ifalias = "blabla" + port.ifalias
        assert portstate_pm.matches_portstate(device, port)

    @pytest.mark.parametrize("portstate_pm", [MatchType.INTF_REGEXP], indirect=True)
    def test_intf_regexp_should_return_true_for_matching_port_and_device(self, portstate_pm, device, port):
        assert portstate_pm.matches_portstate(device, port)

    @pytest.mark.parametrize("portstate_pm", [MatchType.INTF_REGEXP], indirect=True)
    def test_intf_regexp_match_type_should_return_false_for_matching_port_and_non_matching_device(
        self, portstate_pm, device, port
    ):
        device.name = "wrong"
        assert not portstate_pm.matches_portstate(device, port)

    @pytest.mark.parametrize("portstate_pm", [MatchType.EXACT], indirect=True)
    def test_should_return_false_if_invalid_match_type(self, portstate_pm, device, port):
        assert not portstate_pm.matches_portstate(device, port)


@pytest.fixture
def portstate_pm(request, device, port) -> PortStateMaintenance:
    match_type = request.param
    if match_type == MatchType.INTF_REGEXP:
        match_expression = port.ifdescr
    else:
        match_expression = port.ifalias
    return PortStateMaintenance(
        start_time=datetime.datetime.now() - datetime.timedelta(days=1),
        end_time=datetime.datetime.now() + datetime.timedelta(days=1),
        match_type=match_type,
        match_expression=match_expression,
        match_device=device.name,
    )


@pytest.fixture
def matching_portstate_pm(device, port) -> Iterator[PortStateMaintenance]:
    with patch("zino.statemodels.PortStateMaintenance.matches_portstate") as mock:
        mock.return_value = True
        yield PortStateMaintenance(
            start_time=datetime.datetime.now() - datetime.timedelta(days=1),
            end_time=datetime.datetime.now() + datetime.timedelta(days=1),
            match_type=MatchType.STR,
            match_expression=port.ifalias,
            match_device=device.name,
        )


@pytest.fixture
def nonmatching_portstate_pm(device, port) -> Iterator[PortStateMaintenance]:
    with patch("zino.statemodels.PortStateMaintenance.matches_portstate") as mock:
        mock.return_value = False
        yield PortStateMaintenance(
            start_time=datetime.datetime.now() - datetime.timedelta(days=1),
            end_time=datetime.datetime.now() + datetime.timedelta(days=1),
            match_type=MatchType.STR,
            match_expression=port.ifalias,
            match_device=device.name,
        )



import logging
from datetime import timedelta
from ipaddress import IPv4Address
from unittest.mock import AsyncMock, patch

import pytest

from zino.config.models import PollDevice
from zino.state import ZinoState
from zino.statemodels import (
    BGPAdminStatus,
    BGPEvent,
    BGPOperState,
    BGPPeerSession,
    BGPStyle,
)
from zino.tasks.bgpstatemonitortask import BaseBGPRow, BGPStateMonitorTask
from zino.time import now

PEER_ADDRESS = IPv4Address("10.0.0.1")
DEFAULT_REMOTE_AS = 20
DEFAULT_UPTIME = 250


class TestBGPStateMonitorTask:
    @pytest.mark.parametrize("task", ["public", "juniper-bgp", "cisco-bgp", "general-bgp"], indirect=True)
    async def test_task_runs_without_errors(self, task):
        assert (await task.run()) is None

    @pytest.mark.parametrize("task", ["missing-info-bgp"], indirect=True)
    async def test_task_logs_missing_information(self, task, caplog):
        """Tests that the BGP state monitor task logs if necessary information for a BGP device is missing"""
        with caplog.at_level(logging.INFO):
            await task.run()

        assert f"router {task.device.name} misses BGP variables" in caplog.text

    @pytest.mark.parametrize(
        "task", ["general-bgp-external-reset", "cisco-bgp-external-reset", "juniper-bgp-external-reset"], indirect=True
    )
    async def test_external_reset_creates_event(self, task):
        """Tests that an event should be made if a BGP connection to a device has been reset"""
        # set initial state
        task.device_state.bgp_peers = {
            PEER_ADDRESS: BGPPeerSession(uptime=500, admin_status=BGPAdminStatus.STOP, oper_state=BGPOperState.DOWN)
        }
        await task.run()
        # check if state has been updated to reflect state defined in .snmprec
        assert task.device_state.bgp_peers[PEER_ADDRESS].admin_status in [BGPAdminStatus.RUNNING, BGPAdminStatus.START]
        assert task.device_state.bgp_peers[PEER_ADDRESS].oper_state == BGPOperState.ESTABLISHED
        # check that the correct event has been created
        event = task.state.events.get(device_name=task.device.name, subindex=PEER_ADDRESS, event_class=BGPEvent)
        assert event
        assert event.admin_status in [BGPAdminStatus.RUNNING, BGPAdminStatus.START]
        assert event.operational_state == BGPOperState.ESTABLISHED
        assert event.remote_address == PEER_ADDRESS
        assert event.remote_as == DEFAULT_REMOTE_AS
        assert event.peer_uptime == DEFAULT_UPTIME

    @pytest.mark.parametrize(
        "task", ["general-bgp-external-reset", "cisco-bgp-external-reset", "juniper-bgp-external-reset"], indirect=True
    )
    async def test_session_up_updates_event(self, task):
        """Tests that the oper down event should be updated if a BGP connection to a device reports that their
        oper_state has changed back to established
        """
        # create admin down event
        event = task.state.events.get_or_create_event(
            device_name=task.device.name, subindex=PEER_ADDRESS, event_class=BGPEvent
        )
        event.operational_state = BGPOperState.DOWN
        event.admin_status = BGPAdminStatus.STOP
        event.remote_address = PEER_ADDRESS
        event.remote_as = DEFAULT_REMOTE_AS
        event.peer_uptime = 0
        task.state.events.commit(event=event)

        await task.run()

        # check if state has been updated to reflect state defined in .snmprec
        assert task.device_state.bgp_peers[PEER_ADDRESS].admin_status in [BGPAdminStatus.RUNNING, BGPAdminStatus.START]
        assert task.device_state.bgp_peers[PEER_ADDRESS].oper_state == BGPOperState.ESTABLISHED
        # check that the correct event has been created
        event = task.state.events.get(device_name=task.device.name, subindex=PEER_ADDRESS, event_class=BGPEvent)
        assert event
        assert event.admin_status in [BGPAdminStatus.RUNNING, BGPAdminStatus.START]
        assert event.operational_state == BGPOperState.ESTABLISHED
        assert event.remote_address == PEER_ADDRESS
        assert event.remote_as == DEFAULT_REMOTE_AS
        assert event.peer_uptime == DEFAULT_UPTIME

    @pytest.mark.parametrize(
        "task", ["general-bgp-admin-down", "cisco-bgp-admin-down", "juniper-bgp-admin-down"], indirect=True
    )
    async def test_admin_down_creates_event(self, task):
        """Tests that an event should be made if a BGP connection to a device reports that their admin_status has
        changed from start to stop
        """
        # set initial state
        task.device_state.bgp_peers = {
            PEER_ADDRESS: BGPPeerSession(
                uptime=DEFAULT_UPTIME, admin_status=BGPAdminStatus.START, oper_state=BGPOperState.ESTABLISHED
            )
        }
        await task.run()
        # check if state has been updated to reflect state defined in .snmprec
        assert task.device_state.bgp_peers[PEER_ADDRESS].admin_status in [BGPAdminStatus.HALTED, BGPAdminStatus.STOP]
        assert task.device_state.bgp_peers[PEER_ADDRESS].oper_state != BGPOperState.ESTABLISHED
        # check that the correct event has been created
        event = task.state.events.get(device_name=task.device.name, subindex=PEER_ADDRESS, event_class=BGPEvent)
        assert event
        assert event.admin_status in [BGPAdminStatus.HALTED, BGPAdminStatus.STOP]
        assert event.operational_state == BGPOperState.DOWN
        assert event.remote_address == PEER_ADDRESS
        assert event.remote_as == DEFAULT_REMOTE_AS
        assert event.peer_uptime == 0

    @pytest.mark.parametrize(
        "task", ["general-bgp-admin-up", "cisco-bgp-admin-up", "juniper-bgp-admin-up"], indirect=True
    )
    async def test_admin_up_creates_event(self, task):
        """Tests that an event should be made if a BGP connection to a device reports that their admin_status has
        changed from stop to start
        """
        # set initial state
        task.device_state.bgp_peers = {
            PEER_ADDRESS: BGPPeerSession(
                uptime=DEFAULT_UPTIME,
                admin_status=BGPAdminStatus.STOP,
                oper_state=BGPOperState.IDLE,
            )
        }
        # create admin down event
        event = task.state.events.get_or_create_event(
            device_name=task.device.name, subindex=PEER_ADDRESS, event_class=BGPEvent
        )
        event.operational_state = BGPOperState.DOWN
        event.admin_status = BGPAdminStatus.STOP
        event.remote_address = PEER_ADDRESS
        event.remote_as = DEFAULT_REMOTE_AS
        event.peer_uptime = 0
        task.state.events.commit(event=event)

        await task.run()
        # check if state has been updated to reflect state defined in .snmprec
        assert task.device_state.bgp_peers[PEER_ADDRESS].admin_status in [BGPAdminStatus.RUNNING, BGPAdminStatus.START]
        assert task.device_state.bgp_peers[PEER_ADDRESS].oper_state != BGPOperState.ESTABLISHED
        # check that the correct event has been created
        event = task.state.events.get(device_name=task.device.name, subindex=PEER_ADDRESS, event_class=BGPEvent)
        assert event
        assert event.admin_status in [BGPAdminStatus.RUNNING, BGPAdminStatus.START]
        assert event.operational_state == BGPOperState.IDLE
        assert event.remote_address == PEER_ADDRESS
        assert event.remote_as == DEFAULT_REMOTE_AS
        assert event.peer_uptime == 0

    @pytest.mark.parametrize(
        "task", ["general-bgp-oper-down", "cisco-bgp-oper-down", "juniper-bgp-oper-down"], indirect=True
    )
    async def test_oper_down_creates_event(self, task):
        """Tests that an event should be made if a BGP connection to a device reports that their oper_state has changed
        from established to something else
        """
        # set initial state
        task.device_state.bgp_peers = {
            PEER_ADDRESS: BGPPeerSession(
                uptime=DEFAULT_UPTIME, admin_status=BGPAdminStatus.START, oper_state=BGPOperState.ESTABLISHED
            )
        }
        await task.run()
        # check if state has been updated to reflect state defined in .snmprec
        assert task.device_state.bgp_peers[PEER_ADDRESS].oper_state != BGPOperState.ESTABLISHED
        # check that the correct event has been created
        event = task.state.events.get(device_name=task.device.name, subindex=PEER_ADDRESS, event_class=BGPEvent)
        assert event
        assert event.admin_status in [BGPAdminStatus.RUNNING, BGPAdminStatus.START]
        assert event.operational_state == BGPOperState.DOWN
        assert event.remote_address == PEER_ADDRESS
        assert event.remote_as == DEFAULT_REMOTE_AS
        assert event.peer_uptime == 1000000

    @pytest.mark.parametrize(
        "task", ["general-bgp-oper-down", "cisco-bgp-oper-down", "juniper-bgp-oper-down"], indirect=True
    )
    async def test_oper_down_creates_serializable_event(self, task):
        """Tests that an event created from a BGP state change should be serializable by Pydantic without warnings"""
        # set initial state
        task.device_state.bgp_peers = {
            PEER_ADDRESS: BGPPeerSession(
                uptime=DEFAULT_UPTIME, admin_status=BGPAdminStatus.START, oper_state=BGPOperState.ESTABLISHED
            )
        }
        await task.run()
        # check if state has been updated to reflect state defined in .snmprec
        assert task.device_state.bgp_peers[PEER_ADDRESS].oper_state != BGPOperState.ESTABLISHED
        # check that the correct event has been created
        event = task.state.events.get(device_name=task.device.name, subindex=PEER_ADDRESS, event_class=BGPEvent)
        assert event
        assert event.model_dump_json(exclude_none=True, indent=2, warnings="error")

    @pytest.mark.parametrize(
        "task",
        ["general-bgp-oper-down-short", "cisco-bgp-oper-down-short", "juniper-bgp-oper-down-short"],
        indirect=True,
    )
    async def test_oper_down_does_not_create_event_if_uptime_less_than_specified(self, task):
        """Tests that an event should not be made if a BGP connection to a device reports that their oper_state has
        changed from established to something else, but the uptime is less than a specified time
        """
        # set initial state
        if task.device_state.bgp_style == BGPStyle.JUNIPER:
            task.device_state.bgp_peers = {
                PEER_ADDRESS: BGPPeerSession(
                    uptime=DEFAULT_UPTIME, admin_status=BGPAdminStatus.RUNNING, oper_state=BGPOperState.ESTABLISHED
                )
            }
        else:
            task.device_state.bgp_peers = {
                PEER_ADDRESS: BGPPeerSession(
                    uptime=DEFAULT_UPTIME, admin_status=BGPAdminStatus.START, oper_state=BGPOperState.ESTABLISHED
                )
            }
        await task.run()
        # check if state has been updated to reflect state defined in .snmprec
        assert task.device_state.bgp_peers[PEER_ADDRESS].oper_state != BGPOperState.ESTABLISHED
        # check that no event has been created
        event = task.state.events.get(device_name=task.device.name, subindex=PEER_ADDRESS, event_class=BGPEvent)
        assert not event

    @pytest.mark.parametrize(
        "task", ["general-bgp-oper-down", "cisco-bgp-oper-down", "juniper-bgp-oper-down"], indirect=True
    )
    async def test_when_event_is_new_it_should_set_lasttrans(self, task):
        # set initial state
        task.device_state.bgp_peers = {
            PEER_ADDRESS: BGPPeerSession(
                uptime=DEFAULT_UPTIME, admin_status=BGPAdminStatus.START, oper_state=BGPOperState.ESTABLISHED
            )
        }
        await task.run()
        # check that the correct event has been created
        event = task.state.events.get(device_name=task.device.name, subindex=PEER_ADDRESS, event_class=BGPEvent)
        assert event.lasttrans

    @pytest.mark.parametrize(
        "task", ["general-bgp-external-reset", "cisco-bgp-external-reset", "juniper-bgp-external-reset"], indirect=True
    )
    async def test_when_event_transitions_from_not_established_to_established_it_should_update_lasttrans(self, task):
        # create oper down event
        event = task.state.events.get_or_create_event(
            device_name=task.device.name, subindex=PEER_ADDRESS, event_class=BGPEvent
        )
        event.operational_state = BGPOperState.DOWN
        event.admin_status = BGPAdminStatus.STOP
        event.remote_address = PEER_ADDRESS
        event.remote_as = DEFAULT_REMOTE_AS
        event.peer_uptime = 0
        task.state.events.commit(event=event)

        # initial lasttrans should be set by commit()
        initial_lasttrans = event.lasttrans

        await task.run()
        updated_event = task.state.events[event.id]
        assert updated_event.operational_state == BGPOperState.ESTABLISHED
        assert updated_event.lasttrans > initial_lasttrans

    @pytest.mark.parametrize(
        "task", ["general-bgp-admin-down", "cisco-bgp-admin-down", "juniper-bgp-admin-down"], indirect=True
    )
    async def test_when_event_transitions_from_established_to_not_established_it_should_update_lasttrans(self, task):
        initial_lasttrans = now() - timedelta(minutes=5)
        # create oper established event
        event = task.state.events.get_or_create_event(
            device_name=task.device.name, subindex=PEER_ADDRESS, event_class=BGPEvent
        )
        event.operational_state = BGPOperState.ESTABLISHED
        event.admin_status = BGPAdminStatus.RUNNING
        event.remote_address = PEER_ADDRESS
        event.remote_as = DEFAULT_REMOTE_AS
        event.peer_uptime = 0
        event.lasttrans = initial_lasttrans
        task.state.events.commit(event=event)

        await task.run()
        updated_event = task.state.events[event.id]
        assert updated_event.operational_state == BGPOperState.DOWN
        assert updated_event.lasttrans > initial_lasttrans

    @pytest.mark.parametrize(
        "task", ["general-bgp-external-reset", "cisco-bgp-external-reset", "juniper-bgp-external-reset"], indirect=True
    )
    async def test_when_event_transitions_from_not_established_to_established_it_should_update_ac_down(self, task):
        initial_ac_down = timedelta(0)

        # create oper down event
        event = task.state.events.get_or_create_event(
            device_name=task.device.name, subindex=PEER_ADDRESS, event_class=BGPEvent
        )
        event.operational_state = BGPOperState.DOWN
        event.admin_status = BGPAdminStatus.STOP
        event.remote_address = PEER_ADDRESS
        event.remote_as = DEFAULT_REMOTE_AS
        event.peer_uptime = 0
        event.ac_down = initial_ac_down
        task.state.events.commit(event=event)

        await task.run()
        updated_event = task.state.events[event.id]
        assert updated_event.operational_state == BGPOperState.ESTABLISHED
        assert updated_event.ac_down > initial_ac_down


class TestGetBGPStyle:
    @pytest.mark.parametrize("task", ["juniper-bgp"], indirect=True)
    async def test_get_bgp_style_returns_correct_style_for_juniper(self, task):
        assert (await task._get_bgp_style()) == BGPStyle.JUNIPER

    @pytest.mark.parametrize("task", ["cisco-bgp"], indirect=True)
    async def test_get_bgp_style_returns_correct_style_for_cisco(self, task):
        assert (await task._get_bgp_style()) == BGPStyle.CISCO

    @pytest.mark.parametrize("task", ["general-bgp"], indirect=True)
    async def test_get_bgp_style_returns_correct_style_for_general(self, task):
        assert (await task._get_bgp_style()) == BGPStyle.GENERAL

    @pytest.mark.parametrize("task", ["public"], indirect=True)
    async def test_get_bgp_style_returns_correct_style_for_non_bgp_task(self, task):
        assert (await task._get_bgp_style()) is None


class TestGetLocalAs:
    @pytest.mark.parametrize("task", ["juniper-bgp"], indirect=True)
    async def test_get_local_as_returns_correct_value_for_juniper(self, task):
        assert (await task._get_local_as(bgp_style=BGPStyle.JUNIPER)) == 10

    @pytest.mark.parametrize("task", ["cisco-bgp"], indirect=True)
    async def test_get_local_as_returns_correct_value_for_cisco(self, task):
        assert (await task._get_local_as(bgp_style=BGPStyle.CISCO)) == 10

    @pytest.mark.parametrize("task", ["general-bgp"], indirect=True)
    async def test_get_local_as_returns_correct_value_for_general(self, task):
        assert (await task._get_local_as(bgp_style=BGPStyle.GENERAL)) == 10

    @pytest.mark.parametrize("task", ["public"], indirect=True)
    async def test_get_local_as_returns_none_for_non_existent_local_as(self, task):
        assert (await task._get_local_as(bgp_style=BGPStyle.GENERAL)) is None

    @pytest.mark.parametrize("task", ["public"], indirect=True)
    async def test_get_local_as_returns_none_for_non_existent_local_as_with_juniper_bgp_style(self, task):
        assert (await task._get_local_as(bgp_style=BGPStyle.JUNIPER)) is None


class TestBaseBGPRow:
    def test_when_input_is_valid_it_should_not_fail(self):
        assert BaseBGPRow("active", "running", "10.0.42.0", 5, 0)

    def test_when_peer_state_is_invalid_it_should_raise_an_error(self):
        with pytest.raises(ValueError):
            BaseBGPRow("invalid foobar", "running", "10.0.42.0", 5, 0)

    def test_when_peer_admin_status_is_invalid_it_should_raise_an_error(self):
        with pytest.raises(ValueError):
            BaseBGPRow("active", "invalid flimflam", "10.0.42.0", 5, 0)


class TestDisabledBGPDevice:
    async def test_when_do_bgp_is_false_it_should_not_run(self, caplog, task_with_bgp_disabled_device):
        router = task_with_bgp_disabled_device.device.name
        with caplog.at_level(logging.DEBUG):
            await task_with_bgp_disabled_device.run()
            assert f"Skipping BGP scanning for {router} due to config" in caplog.text


@pytest.fixture
def task(request, snmpsim, snmp_test_port):
    device = PollDevice(
        name="buick.lab.example.org",
        address="127.0.0.1",
        community=request.param,
        port=snmp_test_port,
    )
    state = ZinoState()
    task = BGPStateMonitorTask(device, state)
    if BGPStyle.CISCO in request.param:
        task.device_state.bgp_style = BGPStyle.CISCO
    elif BGPStyle.JUNIPER in request.param:
        task.device_state.bgp_style = BGPStyle.JUNIPER
    elif BGPStyle.GENERAL in request.param:
        task.device_state.bgp_style = BGPStyle.GENERAL
    yield task


@pytest.fixture
def task_with_bgp_disabled_device(snmp_test_port):
    device = PollDevice(
        name="buick.lab.example.org",
        address="127.0.0.1",
        community="not-important",
        port=snmp_test_port,
        do_bgp=False,
    )
    state = ZinoState()
    with patch.object(BGPStateMonitorTask, "_get_bgp_style", new=AsyncMock(return_value=BGPStyle.GENERAL)):
        task = BGPStateMonitorTask(device, state)
        yield task



"""SNMP back-end based on netsnmp-cffi"""

import logging
import os
import threading
from collections import defaultdict
from typing import Any, Optional, Sequence, Tuple, Union

from netsnmpy import netsnmp
from netsnmpy.netsnmp import (
    EndOfMibView,
    NoSuchInstance,
    NoSuchObject,
    ObjectIdentifier,
    SNMPVariable,
    symbol_to_oid,
)
from netsnmpy.session import SNMPSession

from zino.config.models import PollDevice
from zino.oid import OID
from zino.snmp.base import (
    EndOfMibViewError,
    Identifier,
    MibNotFoundError,
    MibObject,
    NoSuchInstanceError,
    NoSuchObjectError,
    SNMPBackendError,
    SNMPBackendVersionError,
    SNMPVarBind,
    SparseWalkResponse,
)

_log = logging.getLogger(__name__)


def init_backend():
    """Basic initialization of Net-SNMP library"""
    version = netsnmp.get_version()
    if version < (5, 9):
        raise SNMPBackendVersionError(version)

    if "MIBDIRS" not in os.environ:
        from zino.snmp import get_vendored_mib_directory

        os.environ["MIBDIRS"] = f"+:{get_vendored_mib_directory()}"

    netsnmp.register_log_callback(enable_debug=logging.getLogger("netsnmpy.netsnmp").isEnabledFor(logging.DEBUG))
    netsnmp.load_mibs()
    # Test basic MIB lookup to fail early
    try:
        netsnmp.symbol_to_oid("SNMPv2-MIB::sysUpTime")
        symbol = netsnmp.oid_to_symbol(
            OID(
                ".1.3.6.1.4.1.2636.5.1.1.2.1.1.1.11.0.2.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.2.32.1.7.0.0.0."
                "128.1.0.0.0.0.0.0.0.2"
            )
        )
        assert symbol.startswith("BGP4-V2-MIB-JUNIPER::jnxBgpM2PeerRemoteAddr")
    except (ValueError, AssertionError) as error:
        _log.fatal("MIB tests failed (%s). Make sure the MIBs are loaded correctly.", error)
        _log.fatal("MIBS=%s", os.environ.get("MIBS"))
        _log.fatal("MIBDIRS=%s", os.environ.get("MIBDIRS"))
        raise SNMPBackendError("MIB tests failed. Make sure the MIBs are loaded correctly") from error


class SNMP:
    """Represents an SNMP management session for a single device.

    Net-SNMP allocates low-level resources for each session, so it is important to close the session when done with
    it.  The initial Zino 2 codebase did not expect to have to manage low-level resources, so all SNMP operations of
    this class will implicitly open an unopened or closed session - which means that some parts of Zino *will* have
    to manage resource de-allocation.

    Complete job runs should manage the session as a context manager, but tasks that schedule SNMP operations outside
    of the main job loop should be careful to also manage their own resources.

    Example context manager usage:
    >>> with SNMP(device) as snmp:
    >>>     uptime = await snmp.get("SNMPv2-MIB", "sysUpTime", 0)
    """

    def __init__(self, device: PollDevice):
        self.device = device
        self.session: SNMPSession = SNMPSession(
            host=device.address,
            port=device.port,
            community=device.community,
            version=device.snmpversion,
            timeout=device.timeout,
            retries=device.retries,
        )
        self._is_open = False
        self._lock = threading.RLock()

    def open(self):
        """Opens a low-level SNMP session"""
        if not self._is_open:
            self.session.open()
            self._is_open = True

    def close(self):
        """Closes a low-level SNMP session"""
        if self._is_open:
            self.session.close()
            self._is_open = False

    def __enter__(self):
        self._lock.acquire()
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._lock.release()
        # Silly internal access, but RLock did not gain the `locked()` method until Python 3.14:
        if self._lock._is_owned():
            return  # Only close when no more locks are held
        self.close()

    def _open_if_closed(self):
        if not self._is_open:
            self.open()

    async def get(self, *oid: str) -> MibObject:
        """SNMP-GETs the given oid
        Example usage:
            get("SNMPv2-MIB", "sysUpTime", 0)
            get("1.3.6.1.2.1.1.3.0")

        :param oid: Symbolic MIB object to query.  Multiple formats supported.
        :return: A MibObject representing the resulting MIB variable
        """
        objid = resolve_symbol(oid)
        self._open_if_closed()
        var_binds = await self.session.aget(objid)
        result = var_binds[0]
        self._raise_varbind_errors(result)
        return MibObject(*result)

    async def get2(self, *variables: Sequence[Union[str, int]]) -> list[SNMPVarBind]:
        """SNMP-GETs multiple variables

        Example usage:
            get2(("SNMPv2-MIB", "sysUpTime", 0), ("SNMPv2-MIB", "sysDescr", 0))
            get2(("1.3.6.1.2.1.1.3.0",))

        :param variables: Symbolic MIB object to query.  Multiple formats supported.
        :return: A list of MibObject instances representing the resulting MIB variables
        """
        oids = [resolve_symbol(var) for var in variables]
        self._open_if_closed()
        var_binds = await self.session.aget(*oids)
        return [_convert_snmp_variable(v) for v in var_binds]

    def _raise_varbind_errors(self, var_bind: netsnmp.Variable):
        """Raises a relevant exception if an error has occurred in a varbind"""
        oid, value = var_bind
        if isinstance(value, NoSuchObject):
            raise NoSuchObjectError(f"Could not find object at {oid}")
        if isinstance(value, NoSuchInstance):
            raise NoSuchInstanceError(f"Could not find instance at {oid}")
        if isinstance(value, EndOfMibView):
            raise EndOfMibViewError("Reached end of MIB view")

    async def getnext(self, *oid: str) -> MibObject:
        """SNMP-GETNEXTs the given oid
        Example usage:
            getnext("SNMPv2-MIB", "sysUpTime")
            getnext("1.3.6.1.2.1.1.3")

        :param oid: Symbolic MIB object to query.  Multiple formats supported.
        :return: A MibObject representing the resulting MIB variable
        """
        objid = resolve_symbol(oid)
        self._open_if_closed()
        var_binds = await self.session.agetnext(objid)
        result = var_binds[0]
        self._raise_varbind_errors(result)
        if result.enum_value is not None:
            value = result.enum_value
        else:
            value = result.value
        return MibObject(result.oid, value)

    async def getnext2(self, *variables: Sequence[str]) -> list[SNMPVarBind]:
        """Dispatches a GET-NEXT query for the given variables and returns a result where the fetched variables are
        identified symbolically.

        Example usage:
        >>> await s.getnext2(("IF-MIB", "ifName", "1"), ("IF-MIB", "ifAlias", "1"])
        [(Identifier("IF-MIB", "ifName", OID('.2')), "Gi2/3"),
         (Identifier("IF-MIB", "ifName", OID('.2')), "Uplink to somewhere")]
        >>>

        :param variables: Symbolic MIB objects to query.  Multiple formats supported.
        :return: A sequence of MibObject instances representing the resulting MIB variables
        """
        oids = [resolve_symbol(var) for var in variables]
        self._open_if_closed()
        var_binds = await self.session.agetnext(*oids)
        return [_convert_snmp_variable(v) for v in var_binds]

    async def walk(self, *oid: str) -> list[MibObject]:
        """Uses SNMP-GETNEXT calls to get all objects in the subtree with oid as root
        Example usage:
            walk("IF-MIB", "ifName")
            walk("1.3.6.1.2.1.31.1.1.1.1")

        :param oid: Symbolic MIB object to query.  Multiple formats supported.
        :return: A list of MibObjects representing the resulting MIB variables
        """
        results = []
        root_oid = resolve_symbol(oid)
        current_oid = root_oid
        self._open_if_closed()
        while True:
            response = await self.session.agetnext(current_oid)
            var_bind = response[0]
            current_oid, value = var_bind
            if not root_oid.is_a_prefix_of(current_oid):
                break
            results.append(MibObject(current_oid, value))
        return results

    async def getbulk(self, *oid: str, max_repetitions: int = 1) -> list[MibObject]:
        """SNMP-BULKs the given oid
        Example usage:
            getbulk("IF-MIB", "ifName", max_repetitions=5)
            getbulk("1.3.6.1.2.1.31.1.1.1.1")

        :param oid: Symbolic MIB object to query.  Multiple formats supported.
        :param max_repetitions: Max amount of MIB objects to retrieve
        :return: A list of MibObjects representing the resulting MIB variables
        """
        objid = resolve_symbol(oid)
        self._open_if_closed()
        var_binds = await self.session.agetbulk(objid, max_repetitions=max_repetitions)
        return [MibObject(*var_bind) for var_bind in var_binds]

    async def getbulk2(self, *variables: Sequence[str], max_repetitions: int = 5) -> list[list[SNMPVarBind]]:
        """Issues a GET-BULK request for a set of multiple variables, returning the response in a slightly different
         format than getbulk.

        Example usage:
            >>> snmp = SNMP(...)
            >>> snmp.getbulk2(("IF-MIB", "ifName"), ("IF-MIB", "ifAlias"), max_repetitions=5)
            [[(Identifier(mib='IF-MIB', object='ifAlias', index=OID('.1')), 'Uplink'),
              (Identifier(mib='IF-MIB', object='ifDescr', index=OID('.1')), 'GigabitEthernet0/1')],
             [(Identifier(mib='IF-MIB', object='ifAlias', index=OID('.2')), 'example-sw.example.org'),
              (Identifier(mib='IF-MIB', object='ifDescr', index=OID('.2')), 'GigabitEthernet0/2')]]

        :param variables: Variables to fetch, either as OIDs or symbolic names
        :param max_repetitions: Max amount of MIB objects to retrieve
        :return: A sequence of two-tuples that represent the response varbinds
        """
        oid_objects = [resolve_symbol(v) for v in variables]
        self._open_if_closed()
        var_binds = await self.session.agetbulk(*oid_objects, max_repetitions=max_repetitions)

        query_count = len(variables)
        response_chunks = (var_binds[i : i + query_count] for i in range(0, len(var_binds), query_count))

        return [[_convert_snmp_variable(var_bind) for var_bind in chunk] for chunk in response_chunks]

    async def bulkwalk(self, *oid: str, max_repetitions: int = 5) -> list[MibObject]:
        """Uses SNMP-BULK calls to get all objects in the subtree with oid as root
        Example usage:
            bulkwalk("IF-MIB", "ifName", max_repetitions=5)
            bulkwalk("1.3.6.1.2.1.31.1.1.1.1")

        :param oid: Symbolic MIB object to query.  Multiple formats supported.
        :param max_repetitions: Max amount of MIB objects to retrieve per SNMP-BULK call
        :return: A list of MibObjects representing the resulting MIB variables
        """
        results = []
        start_oid = query_oid = resolve_symbol(oid)
        self._open_if_closed()
        while True:
            response = await self.session.agetbulk(query_oid, max_repetitions=max_repetitions)
            if not response:
                break
            for oid, value in response:
                if not start_oid.is_a_prefix_of(oid):
                    return results
                query_oid = oid
                results.append(MibObject(oid, value))
        return results

    async def sparsewalk(self, *variables: Sequence[str], max_repetitions: int = 5) -> SparseWalkResponse:
        """Bulkwalks and returns a "sparse" table.

        A sparse walk is just a walk operation that returns selected columns of table (or, from multiple tables that
        augment each other).

        Example usage:
            >>> snmp = SNMP(...)
            >>> snmp.sparsewalk(("IF-MIB", "ifName"), ("IF-MIB", "ifAlias"))
            {OID('.1'): {"ifname": "1", "ifAlias": "uplink"},
             OID('.2'): {"ifName": "2", "ifAlias": "next-sw.example.org"}}
        """
        # See section 4.2.3 of RFC 3416 for reference.  This method does not support
        # non-repeaters values.
        query_objects = roots = [resolve_symbol(symbol) for symbol in variables]
        results: dict[OID, dict[str, Any]] = defaultdict(dict)

        self._open_if_closed()
        while True:
            query_objects = await self._sparsewalk_iteration(roots, query_objects, results, max_repetitions)
            if not query_objects:
                break  # Nothing left to query
        return dict(results)

    async def _sparsewalk_iteration(
        self, roots: list[OID], query_objects: list[OID], results: dict[OID, dict[str, Any]], max_repetitions: int
    ) -> Optional[list[OID]]:
        """Performs a single iteration of a sparsewalk operation, updating the results dict with the response values.
        Returns the next set of OIDs to query.
        """
        var_binds = await self.session.agetbulk(*query_objects, max_repetitions=max_repetitions)
        if not var_binds:
            return None

        def is_end_of_scope(root: OID, oid: OID, value: Any) -> bool:
            return not root.is_a_prefix_of(oid) or value == EndOfMibView

        # Integrate response values into result dict by processing in chunks of
        # query count. If we want to support non-repeaters, we need to adjust this.
        query_count = len(query_objects)
        response_chunk = (var_binds[i : i + query_count] for i in range(0, len(var_binds), query_count))
        discards = set()
        last_complete_chunk = ()
        for chunk in response_chunk:
            if len(chunk) < query_count:
                # Ignore incomplete chunks, just query those objects over again
                break
            last_complete_chunk = chunk
            for root, query, response in zip(roots, query_objects, chunk):
                oid, value = response
                if is_end_of_scope(root, oid, value):
                    discards.add(oid)
                    continue

                ident, value = _convert_snmp_variable(response)
                results[ident.index][ident.object] = value

        # Prepare next query based on the contents of the last chunk
        query_objects = [oid for oid, _ in last_complete_chunk if oid not in discards]
        return query_objects

    def is_in_scope(self, entry: MibObject, oid: Tuple[str, str]):
        """Returns if the given MibObject is within the subtree defined by the given OID"""
        root = resolve_symbol(oid)
        return root.is_a_prefix_of(entry.oid)

    async def subtree_is_supported(self, *oid: str) -> bool:
        """Returns if the device has an entry for at least one object within the subtree of the given OID"""
        entry = await self.getnext(*oid)
        return entry and self.is_in_scope(entry=entry, oid=oid)

    @staticmethod
    def _var_bind_to_mibobject(var_bind: tuple[OID, Any]) -> MibObject:
        oid, value = var_bind
        return MibObject(oid=oid, value=value)


def _convert_snmp_variable(variable: SNMPVariable) -> SNMPVarBind:
    """Converts a netsnmp-cffi  SNMPVariable to an Identifier/value pair.

    We don't want to use the netsnmp-cffi types outside of this module, so we convert them to our own types.
    """
    mib, obj, _rest = _split_symbol(variable.symbolic_name)
    # Net-SNMP does a symbolic breakdown of the full OID, so we need to reassemble the bits we care about:
    prefix = netsnmp.symbol_to_oid(f"{mib}::{obj}")
    suffix = variable.oid.strip_prefix(prefix)

    if variable.enum_value is not None:
        value = variable.enum_value
    elif variable.textual_convention == "DisplayString":
        value = variable.value.decode("utf-8")
    else:
        value = variable.value

    return Identifier(mib, obj, suffix), value


def _split_symbol(symbol: str) -> Tuple[str, str, str]:
    """Splits a MIB symbol string into its parts"""
    parts = symbol.split("::", maxsplit=1)
    if len(parts) == 2:
        mib, parts = parts
    else:
        mib = ""
    if "." in parts:
        obj, row_index = parts.split(".", maxsplit=1)
    else:
        obj = parts
        row_index = ""
    return mib, obj, row_index


def mib_value_to_python(value: Any) -> Any:
    """Translator of raw MIB object values to plainer Python objects.

    This is useful in the PySNMP backend, but the translations have already taken
    place in the netsnmp-cffi backend, so this is a no-op that exists only for
    compatibility.
    """
    return value


def resolve_symbol(*symbol: ObjectIdentifier) -> OID:
    """Resolves a symbolic MIB object to an OID"""
    try:
        return symbol_to_oid(*symbol)
    except ValueError:
        # Translate the library's lookup error to Zino's expected exception
        raise MibNotFoundError(symbol)
